# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.17.2
# ---

# %% [markdown]
# # Tutorial 3: Multitable
#

# %% [markdown]
#   In this tutorial, we will execute the avatarization of a multi-table dataset. If you want to know more about how the anonymization is performed, you can read [this page](https://docs.octopize.io/docs/principles/method/multi_table/).
#

# %% [markdown]
# ### Setup

# %%
# %matplotlib inline
import os

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from avatar_yaml.models.schema import LinkMethod

from avatars.manager import Manager
from avatars.models import JobKind

username = os.environ.get("AVATAR_USERNAME", "")
password = os.environ.get("AVATAR_PASSWORD", "")

# %%
manager = Manager()  # or manager = Manager(base_url=https://your-server.com)
# Authenticate with the server
manager.authenticate(username, password)

# %% [markdown]
# ## Loading data

# %% [markdown]
# In this tutorial, we will avatarise data that contains a patient table, a doctor table and a visit table
# - There are 130 patients having at least 1 visit
# - There are 50 doctors, they all did at least one visit
# - There are 300 visits
#
# Note:
# - Each table has a primary key with only unique values
# - Each child table refers to a parent table using a foreign key

# %% [markdown]
# <img src="img/multitable.png" style="height:500px" />

# %%
doctor = pd.read_csv("../fixtures/doctor.csv", sep=",")
patient = pd.read_csv("../fixtures/patient.csv", sep=",")
visit = pd.read_csv("../fixtures/visit.csv", sep=",")

# %%
patient.head()

# %%
doctor.head()

# %%
visit.head()

# %% [markdown]
# ## Upload data
# First we need to upload each table in the database.
# Each id column name MUST be specified in the `primary_key` and `foreign_keys` parameter
#
# Note:
# - Any table need variable that is unique for each rows (primary key)
# - When there is a link between 2 tables, we need to specify the links between them.
# - A parent table has at least a primary key. In our case patient and doctor are parent table.
# - A child table must have a primary key and as many foreign keys as it has parent tables. In our case visit is a child table.
# - An individual level table is a dataframe where each row refers to a UNIQUE physical individual. Privacy metrics assess the re-identification risk of these individuals.
# - A table with no child must be at individual level.

# %%
# First initialize the runner
runner = manager.create_runner(set_name="tutorial_multitable")

runner.add_table(
    "patient",
    patient,
    primary_key="p_id",
    individual_level=True,
)

runner.add_table(
    "doctor",
    doctor,
    primary_key="d_id",
    individual_level=True,
)


runner.add_table(
    "visit",
    visit,
    primary_key="visit_id",
    foreign_keys=["patient_id", "doctor_id"],
    individual_level=False,
)

# %% [markdown]
# ## Parameters setup

# %% [markdown]
# Before proceeding to anonymization, you need to specify the links between each table, as well as the anonymization parameters for each table.
#
# We can choose the method to link a child to its parent table after anonymization :
# - *LINEAR_SUM_ASSIGNMENT*: Assign using the linear sum assignment algorithm.
#     This method is a **good privacy and** utility trade-off. The algorithm consumes lots of resources.
#
# - *MINIMUM_DISTANCE_ASSIGNMENT*: Assign using the minimum distance assignment algorithm.
#     This method assigns the closest child to the parent. It is an acceptable privacy and utility
#     trade-off.
#     This algorithm consumes **less resources** than the linear sum assignment.
#
# - *SENSITIVE_ORIGINAL_ORDER_ASSIGNMENT*: Assign the child to the parent using the original order.
#     **WARNING!!! This method is a HIGH PRIVACY BREACH as it keeps the original order to assign
#     the child to the parent.**
#     This method isn't recommended for small datasets for privacy reasons but consumes **less resources** than the other
#     methods.
#
# Note:
# - A link is a relation between 1 parent table and 1 child table

# %%
runner.add_link(
    parent_table_name="patient",
    child_table_name="visit",
    parent_field="p_id",
    child_field="patient_id",
    method=LinkMethod.LINEAR_SUM_ASSIGNMENT,
)
# Linear sum assignment consumes lots of ressource. Change the method if you have a large dataset.

runner.add_link(
    parent_table_name="doctor",
    child_table_name="visit",
    parent_field="d_id",
    child_field="doctor_id",
    method=LinkMethod.LINEAR_SUM_ASSIGNMENT,
)

# %%
runner.advise_parameters()
runner.update_parameters(
    "patient",
    known_variables=[
        "gender",
        "age",
    ],
)

runner.print_parameters()

# %% [markdown]
# ## Anonymization

# %%
runner.run()

# %%
# Get back avatar tables from the results
# Patient
patient_avatar = runner.shuffled("patient")
# Doctor
doctor_avatar = runner.shuffled("doctor")
# Visit
visit_avatar = runner.shuffled("visit")

# %%
patient_avatar.head()

# %%
doctor_avatar.head()

# %%
visit_avatar.head()

# %% [markdown]
# ## Privacy metric computation
# Similarly to multitable avatarization, privacy metrics calculation requires the specification of one set of parameter per table.

# %%
# verify that the job is finished
runner.get_status(JobKind.privacy_metrics)

# %% [markdown]
# ## Privacy metric results
# The privacy metrics results are computed on multiple tables to verify as many attack scenario as possible.
#
# All types of multi table privacy scenario are further described [here](https://docs.octopize.io/docs/understanding/multi_table/).

# %% [markdown]
# <img src="img/multitable_privacy.png" style="height:500px" />

# %% [markdown]
# ## RAW results

# %%
print("*** Privacy metrics on the patient table***")
for method in runner.privacy_metrics("patient"):
    print(f"Computation type : {method['metadata']['computation_type']}")
    print(f"   Hidden rate : {method['hidden_rate']}")

# %% [markdown]
# Key understandings:
# - **Standalone**: Indicates that the patient table, when considered independently, is protected.
# - **To_top_enriched**: Suggests that combining information from the visit table reduces the level of protection.

# %%
print("*** Privacy metrics on the doctor table***")
for method in runner.privacy_metrics("doctor"):
    print(f"Computation type : {method['metadata']['computation_type']}")
    print(f"   Hidden rate : {method['hidden_rate']}")

# %% [markdown]
# Key understandings:
# - **Standalone**: Indicates that the doctor table, when considered independently, is protected.
# - **To_top_enriched**: Suggests that combining information from the visit table reduces the level of protection.

# %% [markdown]
# This summary helps to quickly identify potential privacy risks associated with the multi-table dataset after avatarization. Here patients and doctors tables are well protected even when enriched with visit data.
#
# If you observe a drop between standalone and enriched results, you might consider adjusting the anonymization parameters or assignment methods used during the avatarization process to enhance privacy protection across linked tables.

# %%
runner.render_privacy_metrics_summary()

# %% [markdown]
# # Utility evaluation

# %% [markdown]
# ## Compute signal metrics on multitable context

# %%
# Check the status of the job
runner.get_status(JobKind.signal_metrics)

# %%
print("*** Signal metrics on the patient table***")
for method in runner.signal_metrics("patient"):
    print(f"Computation type : {method['metadata']['computation_type']}")
    print(f"   Hellinger mean : {method['hellinger_mean']}")

# %% [markdown]
# Key insights:
#
# - **Standalone**:  The avatarized patient table preserves the original distribution effectively.
# - **To_top_enriched**: These metrics indicate if the link between the patient and visit tables is well preserved.

# %%
print("*** Signal metrics on the doctor table***")
for method in runner.signal_metrics("doctor"):
    print(f"Computation type : {method['metadata']['computation_type']}")
    print(f"   Hellinger mean : {method['hellinger_mean']}")

# %% [markdown]
# Key insights:
#
# - **Standalone**: The avatarization of the doctor table preserves the original distribution well.
# - **To_top_enriched**: The link between the doctor and visit table is preserved.

# %%
print("*** Signal metrics on the visit table***")
for method in runner.signal_metrics("visit"):
    print(
        f"Computation type : {method['metadata']['computation_type']} with table {method['metadata']['reference']}"
    )
    if method["metadata"]["computation_type"] == "standalone":
        print(f"   Hellinger mean : {method['hellinger_mean']}")
    print(f"   Correlation difference ratio : {method['correlation_difference_ratio']}")

# %% [markdown]
# Key insights:
#
# - **Standalone**: The avatarization of the visit table preserves the original distribution well.
# - **to_bottom_information_propagated with table doctor**: The correlation between the doctor and visit table is well preserved.
# - **to_bottom_information_propagated with table patient**: The correlation between the patient and visit table is well preserved.

# %% [markdown]
# ## Download Report

# %%
runner.download_report("multitable-report.pdf")

# %% [markdown]
# ## Multivariate comparison
#

# %% [markdown]
# You can see how the avatarization affect the link between original and avatar data.
# You can try with different `LinkMethod` and observe how it impacts the link between tables

# %% [markdown]
# ### Visit x Patient

# %%
AVATAR_COLOR = "#3BD6B0"
ORIGINAL_COLOR = "dimgrey"
map_color = {"original": ORIGINAL_COLOR, "avatar": AVATAR_COLOR}
day_order = ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday"]

doctor_avatar_renamed = doctor_avatar.rename(columns={"age": "age_doctor"})
visit_avatar_flat = visit_avatar.join(doctor_avatar_renamed.set_index("d_id"), on="doctor_id")
visit_avatar_flat = visit_avatar_flat.join(patient_avatar.set_index("p_id"), on="patient_id")


doctor_renamed = doctor.rename(columns={"age": "age_doctor"})
visit_flat = visit.join(doctor_renamed.set_index("d_id"), on="doctor_id")
visit_flat = visit_flat.join(patient.set_index("p_id"), on="patient_id")

# %%
visit_flat["day_visit"] = pd.Categorical(
    visit_flat["day_visit"], categories=day_order, ordered=True
)
visit_flat = visit_flat.sort_values("day_visit")
visit_avatar_flat["day_visit"] = pd.Categorical(
    visit_avatar_flat["day_visit"], categories=day_order, ordered=True
)
visit_avatar_flat = visit_avatar_flat.sort_values("day_visit")
fig, axes = plt.subplots(1, 2, figsize=(12, 4))
sns.countplot(
    data=visit_flat,
    x="day_visit",
    hue="gender",
    ax=axes[0],
    palette=[ORIGINAL_COLOR, "lightgrey"],
)
sns.countplot(
    data=visit_avatar_flat,
    x="day_visit",
    hue="gender",
    ax=axes[1],
    palette=[AVATAR_COLOR, "#9fe9d7"],
)
axes[0].set_title("Original")
axes[1].set_title("Avatar")

# %% [markdown]
# ### Visit x Doctor

# %%
visit_flat["day_visit"] = pd.Categorical(
    visit_flat["day_visit"], categories=day_order, ordered=True
)
visit_flat = visit_flat.sort_values(["day_visit", "exam"])
visit_avatar_flat["day_visit"] = pd.Categorical(
    visit_avatar_flat["day_visit"], categories=day_order, ordered=True
)
visit_avatar_flat = visit_avatar_flat.sort_values(["day_visit", "exam"])
fig, axes = plt.subplots(1, 2, figsize=(12, 4))
sns.boxplot(
    data=visit_flat,
    x="age_doctor",
    y="day_visit",
    ax=axes[0],
    color=ORIGINAL_COLOR,
)
sns.boxplot(
    data=visit_avatar_flat,
    x="age_doctor",
    y="day_visit",
    ax=axes[1],
    color=AVATAR_COLOR,
)
axes[0].set_xlim(30, 70)
axes[1].set_xlim(30, 70)
axes[0].set_title("Original")
axes[1].set_title("Avatar")

# %%
visit_avatar_flat = visit_avatar_flat.sort_values("exam")
visit_flat = visit_flat.sort_values("exam")
fig, axes = plt.subplots(1, 2, figsize=(12, 4))
sns.countplot(
    data=visit_flat,
    x="exam",
    hue="job",
    ax=axes[0],
    palette=[ORIGINAL_COLOR, "lightgrey"],
)
sns.countplot(
    data=visit_avatar_flat,
    x="exam",
    hue="job",
    ax=axes[1],
    palette=[AVATAR_COLOR, "#9fe9d7"],
)
axes[0].set_title("Original")
axes[1].set_title("Avatar")

# %% [markdown]
# ### Patient x Doctor

# %%
sns.kdeplot(data=visit_flat, x="age", y="age_doctor", fill=True, color=ORIGINAL_COLOR, alpha=0.8)
sns.kdeplot(
    data=visit_avatar_flat,
    x="age",
    y="age_doctor",
    fill=True,
    color=AVATAR_COLOR,
    alpha=0.8,
)

# %%

# %%



import pandas as pd
from pathlib import Path

# Define paths
DATA_SOURCE = Path(__file__).parent.parent / "data_source"
INPUT_FILE = DATA_SOURCE / "Emploi_ville.csv"
DATA_PROCESSED = Path(__file__).parent.parent / "data_processed"


def load_and_transform_employment_data(reference_date: str = "2025.09.30") -> pd.DataFrame:
    """
    Load employment data, filter by reference date, and pivot to create
    employment columns by gender and status.

    Args:
        reference_date: The date to filter on (default: "2025.09.30")

    Returns:
        DataFrame with columns:
        - Canton, Commune
        - nb_homme_salarie, nb_femme_salarie
        - nb_homme_non_salarie, nb_femme_non_salarie
        - employ_rate, employ_woman_rate, employ_men_rate
    """
    # Load the CSV file
    df = pd.read_csv(INPUT_FILE)

    # Filter by reference date
    df_filtered = df[df["Date de référence"] == reference_date].copy()

    # Create a combined key for pivoting
    df_filtered["genre_statut"] = df_filtered["Genre"] + "_" + df_filtered["Statut"]

    # Pivot the data to get columns for each combination of Genre and Statut
    df_pivot = df_filtered.pivot_table(
        index=["Canton", "Commune"],
        columns="genre_statut",
        values="Nombre de personnes en emploi",
        aggfunc="sum",
    ).reset_index()

    # Rename columns to match requested format
    column_mapping = {
        "Hommes_Salariés": "nb_homme_salarie",
        "Femmes_Salariés": "nb_femme_salarie",
        "Hommes_Non-salariés": "nb_homme_non_salarie",
        "Femmes_Non-salariés": "nb_femme_non_salarie",
    }
    df_pivot = df_pivot.rename(columns=column_mapping)

    # Calculate total employed by gender
    df_pivot["total_hommes"] = df_pivot["nb_homme_salarie"] + df_pivot["nb_homme_non_salarie"]
    df_pivot["total_femmes"] = df_pivot["nb_femme_salarie"] + df_pivot["nb_femme_non_salarie"]
    df_pivot["total_employes"] = df_pivot["total_hommes"] + df_pivot["total_femmes"]

    # Calculate employment rates (proportion of each category)
    # employ_rate: ratio of salaried to total employed
    df_pivot["employ_rate"] = (
        (df_pivot["nb_homme_salarie"] + df_pivot["nb_femme_salarie"]) / df_pivot["total_employes"]
    ).round(4)

    # employ_woman_rate: proportion of women in total employed
    df_pivot["employ_woman_rate"] = (
        df_pivot["nb_femme_salarie"] / df_pivot["total_femmes"]
    ).round(4)

    # employ_men_rate: proportion of men in total employed
    df_pivot["employ_men_rate"] = (df_pivot["nb_homme_salarie"] / df_pivot["total_hommes"]).round(
        4
    )

    # Select and order final columns
    final_columns = [
        "Canton",
        "Commune",
        "nb_homme_salarie",
        "nb_femme_salarie",
        "nb_homme_non_salarie",
        "nb_femme_non_salarie",
        "employ_rate",
        "employ_woman_rate",
        "employ_men_rate",
        "total_employes",
    ]

    return df_pivot[final_columns]


if __name__ == "__main__":
    # Process the data
    result = load_and_transform_employment_data("2025.09.30")

    # Display results
    print(f"Processed {len(result)} communes")
    print("\nFirst 10 rows:")
    print(result.head(10).to_string(index=False))

    # Save to CSV
    output_file = DATA_PROCESSED / "Emploi_ville_processed.csv"
    result.to_csv(output_file, index=False)
    print(f"\nResults saved to: {output_file}")



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.17.2
# ---

# %% [markdown]
# # Tutorial 2: Parameterizing an avatarization

# %% [markdown]
# In this tutorial, we will learn how to set key parameters of the avatarization and how they impact privacy and utility.

# %% [markdown]
# ## Connection

# %%
import os

import pandas as pd
from avatar_yaml.models.parameters import ExcludeVariablesMethod, ImputeMethod

from avatars.constants import PlotKind
from avatars.manager import Manager
from avatars.models import JobKind

username = os.environ.get("AVATAR_USERNAME", "")
password = os.environ.get("AVATAR_PASSWORD", "")

# %% [markdown]
# Run the following cell if your environment does not have all the listed packages already installed.

# %%
manager = Manager()  # or manager = Manager(base_url=https://your-server.com)
# Authenticate with the server
manager.authenticate(username, password)
# Verify that we can connect to the API server
manager.get_health()

# %% [markdown]
# ## Loading data
#
# This tutorial uses the `iris` dataset, allowing us to run several avatarization without delays.

# %%
df = pd.read_csv("../fixtures/iris.csv")

# %%
df.head()

# %%
# The runner is the object that will be used to upload data to the server and run the avatarization
runner_k2 = manager.create_runner("iris_k2")

# Then upload the data, you can either use a pandas dataframe or a file
runner_k2.add_table("iris", df)

# %% [markdown]
# ## Varying k
#
# The parameter *k* can be used to control the tradeoff between privacy and utility. To increase privacy, we recommend increasing the value of k. Because k is the parameter that also impacts the most the utility, it is recommended to alter it gradually.
#
# We demonstrate here the effect of varying *k*.
#

# %%
# Set k
k = 2
runner_k2.set_parameters("iris", k=k)

runner_k2.run()

# Retrieve selected metric
hidden_rate = runner_k2.privacy_metrics("iris")[0]["hidden_rate"]
local_cloaking = runner_k2.privacy_metrics("iris")[0]["local_cloaking"]
hellinger_distance = runner_k2.signal_metrics("iris")[0]["hellinger_mean"]

print(f"With k={k}, the hidden_rate (privacy) is : {hidden_rate}")
print(f"With k={k}, the local_cloaking (privacy) is : {local_cloaking}")
print(f"With k={k}, the hellinger_distance (utility) is : {hellinger_distance}")

# %%
# Create a new runner to run with a different k
runner_k30 = manager.create_runner("iris_k30")
runner_k30.add_table("iris", "../fixtures/iris.csv")

# Set k
k = 30
runner_k30.set_parameters("iris", k=k)

runner_k30.run()

# Retrieve selected metric
hidden_rate = runner_k30.privacy_metrics("iris")[0]["hidden_rate"]
local_cloaking = runner_k30.privacy_metrics("iris")[0]["local_cloaking"]
hellinger_distance = runner_k30.signal_metrics("iris")[0]["hellinger_mean"]

print(f"With k={k}, the hidden_rate (privacy) is : {hidden_rate}")
print(f"With k={k}, the local_cloaking (privacy) is : {local_cloaking}")
print(f"With k={k}, the hellinger_distance (utility) is : {hellinger_distance}")

# download the projections
original_coord_k_30, avatars_coord_k_30 = runner_k30.projections("iris")

# %% [markdown]
# We observe that we are able to increase the level of privacy by simply increasing *k*. But this is at the expense of the utility.

# %% [markdown]
# ## Visualization of originals and avatars

# %% [markdown]
# By looking at originals and avatars in the projected space, we can observe the area covered by avatars and if it covers the same space as the original data.

# %%
# if you are having issues rendering the plot in your notebook, you can use the following line to open the plot in your browser
# runner_k2.render_plot("iris", PlotKind.PROJECTION_3D, open_in_browser=True)
runner_k2.render_plot("iris", PlotKind.PROJECTION_3D)

# %%
runner_k30.render_plot("iris", PlotKind.PROJECTION_3D)

# %% [markdown]
# We observe that the area covered by avatars generated with a low *k* is much closer to the area covered by original data points. We can also see that with a low *k*, some avatars are close to original points that are isolated. This may pose a risk of re-identification. This explains the drop in privacy level when reducing *k*.
#
# Avatars produced with a large *k* are significantly further away from isolated originals and so ensure their privacy. However **care should be taken in setting *k***  with values that are not too high to prevent a drop in utility level. The drop in utility level is represented by the area covered by avatars being much smaller than the ones of originals.

# %% [markdown]
# ## Other parameters

# %% [markdown]
# ### Column weights
#
# Column weights represent the importance of each variable during the projection process. The higher the value for one variable, the more the individuals will be separated regarding this variable.
#
# By default, all variables are given equal weight of 1, but custom weights can be defined to bias the projection towards some specific variables.

# %%
column_weights = {"variety": 3.0}

# %% [markdown]
# ### Number of components
#
# The number of components represents the number of dimensions to consider for the KNN algorithm. With a low value, computation will mostly be based on well-represented variables in the projection.
#
# By default, `ncp` is set to 5, meaning that the 5 dimensions in the projected space that represent the most the variance on the data are used when computing neighbors.

# %%
ncp = 5

# %% [markdown]
# ### Reduction of highly dimensional data
#
# When dealing with variables with a high number of modalities, the number of dimensions created during the projection increase accordingly. This could lead to situations where the number of dimension is too high compared to the number of records.
#
# We developed a parameter to transform categorical data into a restricted number of continuous dimensions embedding categorical variables in vectors.
#
# By default, `use_categorical_reduction` is set to `False`, meaning that categorical variables are left unprocessed. Furthermore, since categorical variables are converted into 20 dimensions by default, it's recommended to enable this parameter if the total number of modalities in the dataset exceeds 20.

# %%
# Turning on categorical reduction will create 20 numeric columns that embed the categorical columns as vectors
# In this particular example it is conterproductive as it dilutes the signal even more (only three modalities in the original dataset)
use_categorical_reduction = True

# %% [markdown]
# ## Categorical variables with large cardinality

# %% [markdown]
# The anonymization of datasets containing categorical variables with large cardinality is not trivial. We recommend to exclude the variable from the avatarization before re-assigning it by individual similarity (`coordinate_similarity`) or by the original row order (`row_order`). Using row order is more likely to preserve identifying information than coordinate similarity. Privacy metrics must be calculated at the end of the process to confirm that the data generated is anonymous.
#
# Metrics are computed after re-assignment of the excluded variables, so a variable that has been excluded is still anonymized as long as the privacy targets are reached.

# %%
exclude_variable_names = ["variety"]
exclude_variable_method = ExcludeVariablesMethod.COORDINATE_SIMILARITY

# %% [markdown]
# ## Missing data
#
# Missing data is common in datasets and is a property that should be modelled.
#
# The Avatar solution can handle variables with missing data without requiring pre-processing. To do so, an additional variable defining whether a value is missing or not will be temporarily added to the data and the missing values will be temporarily imputed. These variables will be part of the anonymization process.
#
# In the presence of missing values, the last step in the avatarization will be to remove temporary variables and add back missing values.
#
# These parameters allow you to choose the method to impute the missing values. `imputation_method` could be : `fast_knn`, `knn` , `mean` , `mode`, `median`. By default we use the `fast_knn` method.

# %%
imputation_method = ImputeMethod.MEAN

# %% [markdown]
# # Running the avatarization

# %% [markdown]
# The avatarization can now be run with different parameters

# %%
runner = manager.create_runner("iris_multi_params")
runner.add_table("iris", "../fixtures/iris.csv")

runner.set_parameters(
    "iris",
    k=k,
    ncp=ncp,
    column_weights=column_weights,
    use_categorical_reduction=use_categorical_reduction,
    exclude_variable_names=exclude_variable_names,
    exclude_variable_method=exclude_variable_method,
    # imputation_method=imputation_method,
)

runner.run(jobs_to_run=[JobKind.standard])

# %% [markdown]
# We will now observe the impact of the parameters on the projections. We recommend executing this last part of the tutorial several times with different settings.

# %%
original_coord, avatars_coord = runner.projections("iris")

runner.render_plot("iris", PlotKind.PROJECTION_3D)

# %% [markdown]
# ## Get suggestions for a parameter set
#
# If you don't need to refine the parameter set around your usage, you can use the application's recommendations.
#

# %%
runner = manager.create_runner("iris_automated")

runner.add_table("iris", "../fixtures/iris.csv")
runner.advise_parameters()

# See the automated parameters
runner.print_parameters()

# %%
# you can see the table summary read by the server
runner.table_summary("iris")

# %% [markdown]
# You can also update the suggested parameters:

# %%
runner.update_parameters("iris", k=2)
runner.print_parameters()
runner.run()

# %% [markdown]
# ## Compute specific privacy metrics
#
# By configuring the privacy metric parameters, you can simulate specific attack scenarios and adjust the thresholds to suit your use case.

# %% [markdown]
# - **`known_variables`**:
#   List of variables that an attacker is likely to know. This list is used to simulate more realistic attack scenarios, where the attacker only has access to a limited subset of information (typically demographic data). This parameter must be defined alongside the **`target`** and is used to evaluate the **linkability** risk.
#
# - **`target`**:
#   Name of a sensitive variable that an attacker might try to infer. This parameter must be used in conjunction with **`known_variables`** and enables the evaluation of **inference** risk.
#
# - **`quantile_threshold`**:
#   Quantile of the holdout distribution used to define a distance and ratio threshold for computing the closest rate.
#   _Example: If set to 5 (default), any avatar falling below the 5th percentile of the holdout distance distribution is considered at risk._

# %%
runner_privacy = manager.create_runner("iris_multi_params")
runner_privacy.add_table("iris", "../fixtures/iris.csv")

runner_privacy.set_parameters(
    "iris",
    k=10,
    known_variables=["sepal.length", "sepal.width"],
    target="variety",
    quantile_threshold=5,
)

runner_privacy.run(jobs_to_run=[JobKind.standard, JobKind.privacy_metrics])

# Retrieve specific metrics
knowledge_protection_rate = runner_privacy.privacy_metrics("iris")[0]["knowledge_protection_rate"]
linkability_protection_rate = runner_privacy.privacy_metrics("iris")[0][
    "linkability_protection_rate"
]
inference_accuracy_ratio = runner_privacy.privacy_metrics("iris")[0]["inference_accuracy_ratio"]
closest_rate_tunned = runner_privacy.privacy_metrics("iris")[0]["closest_rate"]

print(
    f"the knowledge_protection_rate metric measuring correlation is : {knowledge_protection_rate}"
)
print(
    f"the linkability_protection_rate metric measuring linkability is : {linkability_protection_rate}"
)
print(f"the inference_accuracy_ratio metric measuring inference is : {inference_accuracy_ratio}")
print(f"the closest_rate_tunned metric measuring singling out is : {closest_rate_tunned}")

# %% [markdown]
# *In the next tutorial, we will show how to run an avatarization with multiple tables.*
#

# %% [markdown]
#

# %% [markdown]
#



"""
Generate synthetic job seeker dataset based on metadata description
"""

import pandas as pd
import random
from datetime import datetime, timedelta
from faker import Faker

# Initialize Faker for different locales
fake_fr = Faker("fr_FR")
fake_de = Faker("de_DE")
fake_en = Faker("en_US")


def random_date(start_year=1950, end_year=2005):
    """Generate random date between start_year and end_year"""
    start_date = datetime(start_year, 1, 1)
    end_date = datetime(end_year, 12, 31)
    delta = end_date - start_date
    random_days = random.randint(0, abs(delta.days))
    return (start_date + timedelta(days=random_days)).strftime("%Y%m%d")


def random_recent_date(days_back=730):
    """Generate recent date within last days_back days"""
    today = datetime.now()
    start_date = today - timedelta(days=days_back)
    delta = today - start_date
    random_days = random.randint(0, abs(delta.days))
    return (start_date + timedelta(days=random_days)).strftime("%Y%m%d")


def generate_matricule():
    """Generate 13-digit matricule number"""
    return "".join([str(random.randint(0, 9)) for _ in range(13)])


def generate_luxembourg_communes():
    """Return list of Luxembourg communes"""
    return [
        "Luxembourg",
        "Esch-sur-Alzette",
        "Differdange",
        "Dudelange",
        "Pétange",
        "Sanem",
        "Hésperange",
        "Bettembourg",
        "Strassen",
        "Bertrange",
        "Schifflange",
        "Rodange",
        "Ettelbruck",
        "Diekirch",
        "Wiltz",
        "Echternach",
        "Remich",
        "Grevenmacher",
        "Vianden",
        "Clervaux",
        "Mersch",
        "Capellen",
        "Redange",
        "Mondorf-les-Bains",
        "Walferdange",
    ]


def generate_luxembourg_localities():
    """Return dict of localities with codes"""
    localities = {
        "Luxembourg": "110101",
        "Esch-sur-Alzette": "110201",
        "Differdange": "110301",
        "Dudelange": "110401",
        "Echternach": "110502",
        "Grevenmacher": "110601",
        "Remich": "110701",
        "Ettelbruck": "110801",
        "Diekirch": "110901",
        "Wiltz": "111001",
    }
    return localities


def generate_postal_codes():
    """Generate realistic Luxembourg postal codes"""
    return [str(i) for i in range(1000, 10000, 100)]


def generate_nationalities():
    """Return list of common nationalities in Luxembourg"""
    return [
        "Luxembourgeoise",
        "Portugaise",
        "Française",
        "Italienne",
        "Belge",
        "Allemande",
        "Espagnole",
        "Britannique",
        "Néerlandaise",
        "Syrienne",
        "Érythréenne",
        "Somalienne",
        "Afghane",
        "Roumaine",
        "Polonaise",
    ]


def generate_countries():
    """Return list of countries"""
    return [
        "Luxembourg",
        "Portugal",
        "France",
        "Italie",
        "Belgique",
        "Allemagne",
        "Espagne",
        "Royaume-Uni",
        "Pays-Bas",
        "Syrie",
        "Érythrée",
        "Somalie",
        "Afghanistan",
        "Roumanie",
        "Pologne",
    ]


def generate_isco_codes():
    """Generate ISCO occupation codes"""
    return ["2411", "2412", "1211", "5120", "4110", "2310", "2320", "7231", "9321", "5223"]


def generate_rome_codes():
    """Generate ROME occupation codes with descriptions"""
    rome_codes = {
        "M1607": "Secrétariat",
        "M1203": "Comptabilité",
        "K2111": "Formation professionnelle",
        "D1406": "Management en force de vente",
        "I1308": "Assistance informatique",
        "J1501": "Soins infirmiers",
        "H2901": "Ajustement et montage",
        "N1103": "Magasinage et préparation de commandes",
        "K1207": "Intervention socioéducative",
        "G1603": "Personnel polyvalent en restauration",
    }
    return rome_codes


def generate_rome_appellations():
    """Generate ROME occupation appellations"""
    appellations = {
        "M1607": ["Secrétaire", "Secrétaire juridique", "Assistant administratif"],
        "M1203": ["Comptable", "Assistant comptable", "Aide-comptable"],
        "K2111": ["Formateur", "Formateur professionnel", "Conseiller en formation"],
        "D1406": ["Commercial", "Responsable commercial", "Chef des ventes"],
        "I1308": ["Technicien support", "Technicien helpdesk", "Support informatique"],
        "J1501": ["Infirmier", "Aide-soignant", "Auxiliaire de vie"],
        "H2901": ["Monteur", "Ajusteur-monteur", "Mécanicien monteur"],
        "N1103": ["Magasinier", "Préparateur de commandes", "Agent logistique"],
        "K1207": ["Éducateur", "Éducateur spécialisé", "Animateur social"],
        "G1603": ["Serveur", "Employé de restauration", "Agent de restauration"],
    }
    return appellations


def generate_diplomas():
    """Generate diploma types"""
    return [
        "DAP",
        "DT",
        "CATP",
        "Secondaire technique",
        "Secondaire classique",
        "Brevet de maîtrise",
        "BTS",
        "Bachelor",
        "Master",
        "PhD",
        "Diplôme d'éducateur",
    ]


def generate_specialties():
    """Generate education specialties"""
    return [
        "Comptabilité",
        "Économie",
        "Division administrative et commerciale",
        "Informatique",
        "Électronique",
        "Mécanique",
        "Commerce",
        "Hôtellerie-Restauration",
        "Soins infirmiers",
        "Éducation sociale",
        "Droit",
        "Sciences humaines",
        "Langues",
        "Sciences naturelles",
    ]


def generate_titre_sejour():
    """Generate residence permit codes"""
    codes = {
        "065": "Membre de famille (UE avec hors UE)",
        "100": "Protection internationale (BPI)",
        "109": "Protection temporaire (BPT)",
        "120": "Carte de séjour UE",
        "130": "Autorisation de séjour temporaire",
        "": "",  # No permit needed
    }
    return codes


def generate_individual():
    """Generate one individual record"""

    # Basic personal information
    genre = random.choice(["Masculin", "Féminin"])
    if genre == "Masculin":
        nom = fake_fr.last_name()
        prenom = fake_fr.first_name_male()
    else:
        nom = fake_fr.last_name()
        prenom = fake_fr.first_name_female()

    matricule = generate_matricule()
    date_naissance = random_date(1950, 2005)

    # Nationality and birth place
    nationalite = random.choice(generate_nationalities())
    pays_naissance = random.choice(generate_countries())

    # For Luxembourg-born, use Luxembourg communes, otherwise use Faker
    if pays_naissance == "Luxembourg":
        lieu_naissance = random.choice(generate_luxembourg_communes())
    else:
        lieu_naissance = fake_fr.city()

    # Residence information
    type_residence = random.choice(["résident", "résident", "résident", "non-résident", "inconnu"])

    residence_num_rue = str(random.randint(1, 250))
    residence_rue = fake_fr.street_name()
    residence_code_postal = random.choice(generate_postal_codes())

    localities = generate_luxembourg_localities()
    residence_commune = random.choice(list(localities.keys()))
    residence_localite = f"{localities[residence_commune]} {residence_commune}"

    residence_pays = random.choice(
        ["Luxembourg", "Luxembourg", "Luxembourg", "France", "Belgique", "Allemagne"]
    )

    # Civil status and children
    etat_civil = random.choice(
        ["Célibataire", "Marié", "Divorcé", "Veuf", "Séparé", "Inconnu", "PA"]
    )
    enfants = random.choice(["Oui", "Non", "Inconnu"])

    # Special statuses
    statut_sh = random.choice([True, False, False, False, False])
    orientation_sh = (
        random.choice(["Marché ordinaire", "Atelier protégé", "Inconnu"])
        if statut_sh
        else "Inconnu"
    )
    statut_ctr = random.choice([True, False, False, False, False])
    revis = random.choice([True, False, False, False, False])

    # Residence permit
    titre_sejour_dict = generate_titre_sejour()
    has_titre = random.choice([True, False, False, False])
    if has_titre and nationalite not in ["Luxembourgeoise"]:
        titre_sejour_code = random.choice(["065", "100", "109", "120", "130"])
        titre_sejour = f"{titre_sejour_code} {titre_sejour_dict[titre_sejour_code]}"
        debut_titre_sejour = random_recent_date(1095)
        fin_titre_sejour = random_recent_date(-365)  # Future date
    else:
        titre_sejour = ""
        debut_titre_sejour = ""
        fin_titre_sejour = ""

    # Work authorization
    needs_work_auth = random.choice([True, False, False, False])
    if needs_work_auth:
        autorisation_travail_debut = random_recent_date(730)
        autorisation_travail_fin = random_recent_date(-180)
        autorisation_travail_isco = random.choice(generate_isco_codes())
    else:
        autorisation_travail_debut = ""
        autorisation_travail_fin = ""
        autorisation_travail_isco = ""

    # Targeted occupations
    rome_codes = list(generate_rome_codes().keys())
    rome_appellations = generate_rome_appellations()

    metier_rome_1 = random.choice(rome_codes)
    metier_rome_1_appellation = random.choice(rome_appellations[metier_rome_1])
    metier_rome_1_experience = f"{random.randint(0, 20)} ans" if random.random() > 0.3 else ""

    # Second occupation (optional)
    if random.random() > 0.5:
        metier_rome_2 = random.choice([c for c in rome_codes if c != metier_rome_1])
        metier_rome_2_appellation = random.choice(rome_appellations[metier_rome_2])
        metier_rome_2_experience = f"{random.randint(0, 15)} ans" if random.random() > 0.4 else ""
    else:
        metier_rome_2 = ""
        metier_rome_2_appellation = ""
        metier_rome_2_experience = ""

    # Third occupation (optional)
    if random.random() > 0.7:
        metier_rome_3 = random.choice(
            [c for c in rome_codes if c not in [metier_rome_1, metier_rome_2]]
        )
        metier_rome_3_appellation = random.choice(rome_appellations[metier_rome_3])
        metier_rome_3_experience = f"{random.randint(0, 10)} ans" if random.random() > 0.5 else ""
    else:
        metier_rome_3 = ""
        metier_rome_3_appellation = ""
        metier_rome_3_experience = ""

    # Education
    formation_reussie_isced = random.choice(["0", "1", "2", "3", "4", "5", "6", "7", "8"])

    # Diploma based on ISCED level
    if formation_reussie_isced in ["0", "1", "2"]:
        formation_reussie_diplome = random.choice(
            ["", "Secondaire technique", "Secondaire classique"]
        )
        formation_reussie_classe = random.choice(
            ["9ème EST", "12ème EST", "2ème ESC", "5ème primaire"]
        )
    elif formation_reussie_isced == "3":
        formation_reussie_diplome = random.choice(["DAP", "DT", "CATP"])
        formation_reussie_classe = ""
    elif formation_reussie_isced == "4":
        formation_reussie_diplome = random.choice(["Brevet de maîtrise", "Diplôme d'éducateur"])
        formation_reussie_classe = ""
    elif formation_reussie_isced == "5":
        formation_reussie_diplome = "BTS"
        formation_reussie_classe = ""
    elif formation_reussie_isced == "6":
        formation_reussie_diplome = "Bachelor"
        formation_reussie_classe = ""
    elif formation_reussie_isced == "7":
        formation_reussie_diplome = "Master"
        formation_reussie_classe = ""
    else:  # 8
        formation_reussie_diplome = "PhD"
        formation_reussie_classe = ""

    specialties = generate_specialties()
    formation_reussie_specialite = random.choice(specialties) if random.random() > 0.2 else ""
    formation_toutes_specialites = ", ".join(random.sample(specialties, k=random.randint(1, 3)))
    formation_reussie_pays = random.choice(generate_countries())
    formation_reussie_preuve = random.choice(
        ["Déclaré", "Vérifié", "Homologué", "Reconnu", "A valider"]
    )

    # Languages
    langue_fr = random.choice(["A", "B", "C", "C", "B"])  # French more common at higher levels
    langue_lb = random.choice(["", "A", "B", "C"])
    langue_de = random.choice(["", "A", "B", "C"])
    langue_en = random.choice(["", "A", "B", "C"])

    other_langs = ["Portugais", "Italien", "Espagnol", "Néerlandais", "Arabe", "Russe"]
    if random.random() > 0.6:
        langues_autres = f"{random.choice(other_langs)}: {random.choice(['A', 'B', 'C'])}"
    else:
        langues_autres = ""

    # Mobility
    mobilite = random.choice(
        [
            "Excellente (permis A, B et véhicule personnel)",
            "Moyenne (permis B mais pas de voiture disponible ou utilisation de transports en communs)",
            "Limitée (pas de voiture et pas de transports en commun disponible)",
            "Limitée pour des raisons psychiques et/ou physiques (durée de trajet restreinte)",
        ]
    )

    return {
        "Matricule": matricule,
        "Nom": nom,
        "Prenom": prenom,
        "Genre": genre,
        "Date_Naissance": date_naissance,
        "Pays_Naissance": pays_naissance,
        "Lieu_Naissance": lieu_naissance,
        "Nationalite": nationalite,
        "Type_Residence": type_residence,
        "Residence_Num_Rue": residence_num_rue,
        "Residence_Rue": residence_rue,
        "Residence_Code_Postal": residence_code_postal,
        "Residence_Localite": residence_localite,
        "Residence_Commune": residence_commune,
        "Residence_Pays": residence_pays,
        "Etat_Civil": etat_civil,
        "Enfants": enfants,
        "Statut_SH": statut_sh,
        "Orientation_SH": orientation_sh,
        "Statut_CTR": statut_ctr,
        "REVIS": revis,
        "Titre_Sejour": titre_sejour,
        "Debut_Titre_Sejour": debut_titre_sejour,
        "Fin_Titre_Sejour": fin_titre_sejour,
        "Autorisation_Travail_Debut": autorisation_travail_debut,
        "Autorisation_Travail_Fin": autorisation_travail_fin,
        "Autorisation_Travail_ISCO": autorisation_travail_isco,
        "Metier_ROME_1": metier_rome_1,
        "Metier_ROME_1_Appellation": metier_rome_1_appellation,
        "Metier_ROME_1_Experience": metier_rome_1_experience,
        "Metier_ROME_2": metier_rome_2,
        "Metier_ROME_2_Appellation": metier_rome_2_appellation,
        "Metier_ROME_2_Experience": metier_rome_2_experience,
        "Metier_ROME_3": metier_rome_3,
        "Metier_ROME_3_Appellation": metier_rome_3_appellation,
        "Metier_ROME_3_Experience": metier_rome_3_experience,
        "Formation_Reussie_ISCED": formation_reussie_isced,
        "Formation_Reussie_Diplome": formation_reussie_diplome,
        "Formation_Reussie_Classe": formation_reussie_classe,
        "Formation_Reussie_Specialite": formation_reussie_specialite,
        "Formation_Toutes_Specialites": formation_toutes_specialites,
        "Formation_Reussie_Pays": formation_reussie_pays,
        "Formation_Reussie_Preuve": formation_reussie_preuve,
        "Langue_FR": langue_fr,
        "Langue_LB": langue_lb,
        "Langue_DE": langue_de,
        "Langue_EN": langue_en,
        "Langues_Autres": langues_autres,
        "Mobilite": mobilite,
    }


def generate_dataset(n_individuals=200):
    """Generate complete dataset with n individuals"""
    print(f"Generating dataset with {n_individuals} individuals...")

    data = []
    for i in range(n_individuals):
        if (i + 1) % 50 == 0:
            print(f"Generated {i + 1}/{n_individuals} individuals...")
        data.append(generate_individual())

    df = pd.DataFrame(data)
    print(f"Dataset generated successfully with {len(df)} rows and {len(df.columns)} columns")
    return df


if __name__ == "__main__":
    # Generate dataset
    df = generate_dataset(200)

    # Save to CSV
    output_file = "data_processed/job_seekers_synthetic_200.csv"
    df.to_csv(output_file, index=False, encoding="utf-8-sig")
    print(f"\nDataset saved to: {output_file}")

    # Display summary
    print("\n=== Dataset Summary ===")
    print(f"Total individuals: {len(df)}")
    print(f"Total columns: {len(df.columns)}")
    print(f"\nFirst few rows:")
    print(df.head())
    print(f"\nColumn names:")
    print(df.columns.tolist())



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.17.2
# ---

# %% [markdown]
# # Tutorial 5: Client-side processors

# %% [markdown]
# In this tutorial, we will learn how to prepare the data prior to running an avatarization by using processors on your local machine.
#
# This step is necessary in some cases to handle and preserve data characteristics that are not natively handled by the avatarization or its embedded processors.
#
# We'll also show how custom client-side processors can be defined to integrate domain knowledge into an avatarization.

# %% [markdown]
# ## Principles

# %% [markdown]
# ![pipeline](img/pipeline.png)

# %% [markdown]
# ## Connection

# %%
import os

import numpy as np
import pandas as pd

from avatars.manager import Manager
from avatars.models import JobKind
from avatars.processors import (
    ExpectedMeanProcessor,
    GroupModalitiesProcessor,
    PerturbationProcessor,
    ProportionProcessor,
    RelativeDifferenceProcessor,
)

username = os.environ.get("AVATAR_USERNAME", "")
password = os.environ.get("AVATAR_PASSWORD", "")

# %%
manager = Manager()  # or manager = Manager(base_url=https://your-server.com)
# Authenticate with the server
manager.authenticate(username, password)
# Verify that we can connect to the API server
manager.get_health()

# %% [markdown]
# ## A helper processor to reduce the number of modalities

# %% [markdown]
# We have seen in the previous tutorial one approach to handle categorical variables with large cardinality. We propose here an alternative way of doing this using a client-side processor.
#
# This processor will group modalities together to ensure the target variable has a requested number of modalities. The least represented modalities will be brought together under a `other` modality. Note that this transformation is irreversible (the original value cannot be infered from `other`).
#
# Because this is an irreversible operation, this transformation of the data should be done outside the pipeline. The transformed data will be used as a basis for comparison when computing utility and privacy metrics.

# %%
df = pd.read_csv("../fixtures/adult_with_cities.csv").head(1000).drop(["native-country"], axis=1)
df.head()

# %% [markdown]
# After loading the data, we decide we wish to reduce the number of modalities for the variable `city` which contains originally over 80 distinct values.

# %%
df["city"].value_counts()

# %%
group_modalities_processor = GroupModalitiesProcessor(
    min_unique=10,  # number of modalities for a variable to be considered for grouping
    global_threshold=25,  # if considered for grouping, number of individuals in modality to preserve it
    new_category="other",
)
df_preprocessed = group_modalities_processor.preprocess(df)

# %% [markdown]
# Once the group modality processor has been applied, we can confirm that the number of modalities for the `city` variables has been reduced

# %%
df_preprocessed["city"].value_counts()

# %%
runner = manager.create_runner(set_name="tutorial5")
runner.add_table(
    "adult",
    df_preprocessed,
)
runner.set_parameters("adult", k=5)
runner.run(jobs_to_run=[JobKind.standard])
runner.get_all_results()

# %%
avatars = runner.shuffled("adult")
avatars.head(3)

# %%
avatars["city"].value_counts()

# %% [markdown]
# We observe that the avatars produced have a reduced number of cities and an extra `other` modality for the `city` variable. Note that the use of a client-side processor made the transformation of the data straightforward.

# %% [markdown]
# ## Modeling inter-variables constraints with processors
# We will now use two processors to enforce inter-variable constraints.
#
# The two processors we will now apply are processors that temporarily transform the data in order to improve the avatarization. This means that they both contain a `preprocess` step and a `postprocess` step, ensuring that the effect of the `preprocess` action can be reversed via the use of the `postprocess` action.
#
# These processors will be used to demonstrate the implementation of the pipeline tool that automates the use of processors, the avatarization and the metric computation in a single command.
#
#

# %%
df = pd.read_csv("../fixtures/epl.csv")

# %% [markdown]
# ### Proportions
# Variables may have relationships in which one or many variables can be represented as a proportion of another. In order to best preserve this type of relationships during avatarization, it is recommended to express such variables as proportions. To do so, the proportion processor can be used.
# In our example, the `ProportionProcessor` can be used, because `minutes_in_game` is the sum of `minutes_played_home`, `minutes_played_away` and `minutes_on_bench`
#
# $\displaystyle\frac{minutes\,in\,game\:+\:minutes\,played\,away\:+\:minutes\,on\,bench}{ minutes\,in\,game} = 1$
#

# %%
proportion_processor = ProportionProcessor(
    variable_names=["minutes_played_home", "minutes_played_away", "minutes_on_bench"],
    reference="minutes_in_game",
    sum_to_one=True,
    decimal_count=0,
)

# %% [markdown]
# ### Relative differences

# %% [markdown]
# Some variables may have a hierarchy where one variable is always higher than an other. In order to be sure that this hierarchy is preserved in the avatarization, it is recommended to express one variable as the difference of another.
# We take `penalty_attempts` and `penalty_goals` as an example where one variable (`penalty_goals`) cannot be greater than the other (`penalty_attempts`).

# %%
relative_difference_processor = RelativeDifferenceProcessor(
    target="penalty_goals",
    references=["penalty_attempts"],
)

# %% [markdown]
# ### Computed variables
# The data also contains a third variable related to the penalty context: `penalty_misses`. This variable can be computed directly as the difference between `penalty_attempts` and `penalty_goals`.
#
# $\ penalty\:misses = penalty\:attempts - penalty\:goals $
#
# Computed variables should be removed from the data prior to running the avtarization and re-computed after.
#

# %%
df = df.drop(columns=["penalty_misses"])

# %%
preprocessed = proportion_processor.preprocess(df)
preprocessed = relative_difference_processor.preprocess(preprocessed)

# %%
runner_with_processor = manager.create_runner(set_name="tutorial5")
runner_with_processor.add_table(
    "game",
    preprocessed,
)
runner_with_processor.set_parameters("game", k=5, use_categorical_reduction=True)
runner_with_processor.run()
runner_with_processor.get_all_results()

# %%
postprocessed = proportion_processor.postprocess(df, runner_with_processor.shuffled("game"))
postprocessed = relative_difference_processor.postprocess(
    df, runner_with_processor.shuffled("game")
)

# %%
privacy_metrics = runner_with_processor.privacy_metrics("game")[0]
print("*** Privacy metrics ***")
for metric in privacy_metrics:
    print(metric)

# %%
signal_metrics = runner_with_processor.signal_metrics("game")[0]
print("*** Signal metrics ***")
for metric in signal_metrics:
    print(metric)

# %% [markdown]
# ### Should these processors really be used ?
# Let's try without ...

# %%
runner_no_processor = manager.create_runner(set_name="tutorial5")
runner_no_processor.add_table(
    "game",
    df,
)
runner_no_processor.set_parameters("game", k=5, use_categorical_reduction=True)
runner_no_processor.run(jobs_to_run=[JobKind.standard])
runner_no_processor.get_all_results()

# %% [markdown]
# #### Preservation of the proportions
# To confirm that proportions are well kept, we can compute the maximum difference between the reference variable (`minutes_in_game`) and the sum of the three proportion variables (`minutes_played_home`, `minutes_played_away` and `minutes_on_bench`). Where it may not be zero when no processor is used, this difference should be zero when using a proportion processor.

# %%
avatars_noprocessing = runner_no_processor.shuffled("game")
np.max(
    abs(
        avatars_noprocessing["minutes_in_game"]
        - (
            avatars_noprocessing["minutes_played_home"]
            + avatars_noprocessing["minutes_played_away"]
            + avatars_noprocessing["minutes_on_bench"]
        )
    )
)

avatars_processing = runner_with_processor.shuffled("game")
np.max(
    abs(
        avatars_processing["minutes_in_game"]
        - (
            avatars_processing["minutes_played_home"]
            + avatars_processing["minutes_played_away"]
            + avatars_processing["minutes_on_bench"]
        )
    )
)

# %% [markdown]
# #### Preservation of the relative difference
#
# To confirm that the relative difference is preserved, we can look at the number of players who have more `penalty_goals` than `penalty_attempts`. With the `RelativeDifferenceProcessor` zero players should have more goals than attempts, which is not necessarily the case without processor.

# %%
print("Avatars with processors")
print(
    "Number of players with penalty attempts > penalty goals: ",
    (sum(avatars_processing["penalty_attempts"] - avatars_processing["penalty_goals"] > 0)),
)
print(
    "Number of players with penalty attempts < penalty goals: ",
    (sum(avatars_processing["penalty_attempts"] - avatars_processing["penalty_goals"] < 0)),
)

print("Avatars without processors")
print(
    "Number of players with penalty attempts > penalty goals: ",
    (sum(avatars_noprocessing["penalty_attempts"] - avatars_noprocessing["penalty_goals"] > 0)),
)
print(
    "Number of players with penalty attempts < penalty goals: ",
    (sum(avatars_noprocessing["penalty_attempts"] - avatars_noprocessing["penalty_goals"] < 0)),
)

# %% [markdown]
# ## Post-processors
#
# Post-processors do not transform the data prior to the avatarization but after only. These can be used to fix some variables that could have been altered beyond acceptable.
#
# Care should always be taken when using post-processors because they are likely to **decrease the level of privacy**.
# Using these processors via the pipeline feature, ensure that metrics are computed after application of the post-process step. This also make sure that privacy and utility metrics have taken these processors into consideration.
#

# %% [markdown]
# ### Expected mean
#
# `ExpectedMeanProcessor` is used to force values to have similar mean to original data. In our example, we want to preserve the mean of `goals_away` and `goals_home` by the variable `position`.
# Care should be taken when using this processor as it only targets enhancement of unimodal utility. This may occur at the expense of multi-modal utility and **privacy**.
#

# %%
expected_mean_processor = ExpectedMeanProcessor(
    target_variables=["goals_away", "goals_home"],
    groupby_variables=["position"],
    same_std=False,
)
preprocessed = expected_mean_processor.preprocess(df)

# %%
runner = manager.create_runner(set_name="tutorial5")
runner.add_table(
    "game",
    preprocessed,
)
runner.set_parameters("game", k=5, use_categorical_reduction=True)
runner.run(jobs_to_run=[JobKind.standard])
runner.get_all_results()

# %% [markdown]
# Looking at the mean of the two variables on which the expected mean processor was applied, we can confirm that the mean for each target category is preserved.
# The same statistics computed on avatars that did not get post-processed are more different than the statistics obtained on the original data.

# %%
df.groupby(["position"]).mean(numeric_only=True)[["goals_away", "goals_home"]]

# %%
avatars_processing.groupby(["position"]).mean(numeric_only=True)[["goals_away", "goals_home"]]

# %%
avatars_noprocessing.groupby(["position"]).mean(numeric_only=True)[["goals_away", "goals_home"]]

# %% [markdown]
# ### Computed variables
#
# To complete the anonymization process, computed variables (that have been removed from the data) should be added back to the avatarized data.

# %%
avatars_processing["penalty_missed"] = (
    avatars_processing["penalty_attempts"] - avatars_processing["penalty_goals"]
)
avatars_noprocessing["penalty_missed"] = (
    avatars_noprocessing["penalty_attempts"] - avatars_noprocessing["penalty_goals"]
)

# %% [markdown]
# ### Perturbation level
#
# The perturbation processor can be used to control how close to the avatarized values, the final values of a variable will be. At the extremes, if using a perturbation level of zero, the avatarized values will not contribute at all to the final values. On the other hand, with a perturbation level of 1, the original values will not contribute. A perturbation level of 0.3 will mean that the final value will be closer to the original values than it is from the anonymized values. By default, the perturbation level is set to 1.
#

# %%
perturbation_processor = PerturbationProcessor(perturbation_level={"age": 0, "appearances": 1})
preprocessed = perturbation_processor.preprocess(df)

# %%
runner = manager.create_runner(set_name="tutorial5")
runner.add_table(
    "game",
    preprocessed,
)
runner.set_parameters("game", k=5, use_categorical_reduction=True)
runner.run(jobs_to_run=[JobKind.standard])
runner.get_all_results()
avatars_perturbation = runner.shuffled("game")

# %%
postprocessed = perturbation_processor.postprocess(df, runner.shuffled("game"))

# %% [markdown]
# We observe that as expected, using a perturbation level of 0 on the variable `age`, this variable gets unchanged.

# %%
df["age"].value_counts() - postprocessed["age"].value_counts()

# %% [markdown]
# The same comment does not hold when using a perturbation level of 1.
# A count of the number of modalities shows that new modalities were created during avatarization

# %%
df["appearances"].value_counts() - postprocessed["appearances"].value_counts()



"""
Synthetic Citizen (Job Seeker) Dataset Generator

Generates realistic synthetic data for job seekers ("demandeurs d'emploi")
respecting the metadata rules defined in metadata.csv.

The generator implements realistic correlations:
- Age correlates with marital status, education, experience
- Education correlates with job type and salary expectations
- Nationality correlates with language skills
- Residence type affects work authorization requirements
- Education level affects mobility and job targets
"""

import random
from datetime import date, timedelta
from pathlib import Path
from typing import Any

import numpy as np
import pandas as pd

# Seed for reproducibility
random.seed(42)
np.random.seed(42)

DATA_SOURCE = Path(__file__).parent.parent / "data_source"
DATA_PROCESSED = Path(__file__).parent.parent / "data_processed"

# ============================================================================
# REFERENCE DATA
# ============================================================================

# Luxembourg communes with postal codes
COMMUNES_LU = [
    ("Luxembourg", "1000"),
    ("Esch-sur-Alzette", "4000"),
    ("Differdange", "4500"),
    ("Dudelange", "3400"),
    ("Pétange", "4700"),
    ("Ettelbruck", "9000"),
    ("Diekirch", "9200"),
    ("Strassen", "8000"),
    ("Bertrange", "8100"),
    ("Bettembourg", "3200"),
    ("Hesperange", "5765"),
    ("Schifflange", "3800"),
    ("Sanem", "4400"),
    ("Mamer", "8200"),
    ("Mersch", "7500"),
    ("Kayl", "3600"),
    ("Rumelange", "3700"),
    ("Echternach", "6400"),
    ("Grevenmacher", "6700"),
    ("Remich", "5500"),
    ("Mondorf-les-Bains", "5600"),
    ("Clervaux", "9700"),
    ("Vianden", "9400"),
    ("Wiltz", "9500"),
]

# Neighboring countries (for non-residents / cross-border workers)
CROSS_BORDER = {
    "France": [
        ("Thionville", "57100"),
        ("Metz", "57000"),
        ("Longwy", "54400"),
        ("Audun-le-Tiche", "57390"),
    ],
    "Belgique": [
        ("Arlon", "6700"),
        ("Virton", "6760"),
        ("Aubange", "6790"),
    ],
    "Allemagne": [
        ("Trier", "54290"),
        ("Bitburg", "54634"),
        ("Saarburg", "54439"),
    ],
}

# Nationalities with weights (reflecting Luxembourg demographics)
NATIONALITIES = {
    "Luxembourgeoise": 0.30,
    "Portugaise": 0.16,
    "Française": 0.15,
    "Italienne": 0.04,
    "Belge": 0.04,
    "Allemande": 0.03,
    "Espagnole": 0.02,
    "Britannique": 0.01,
    "Polonaise": 0.02,
    "Roumaine": 0.02,
    "Cap-Verdienne": 0.01,
    "Marocaine": 0.01,
    "Ukrainienne": 0.02,
    "Syrienne": 0.01,
    "Érythréenne": 0.01,
    "Autres": 0.15,
}

# Common first names by gender (Luxembourg/European mix)
FIRST_NAMES_M = [
    "Jean", "Pierre", "Marc", "Michel", "Paul", "André", "Claude", "François",
    "Luc", "Patrick", "Thierry", "Carlos", "José", "Manuel", "António",
    "Marco", "Fabio", "Matteo", "Stefan", "Thomas", "Alexander", "Sébastien",
    "Nicolas", "Philippe", "Laurent", "David", "Julien", "Romain", "Yannick",
    "Piotr", "Andrei", "Oleksandr", "Mohammed", "Ahmed", "Ali",
]

FIRST_NAMES_F = [
    "Marie", "Anne", "Jeanne", "Sophie", "Claire", "Isabelle", "Catherine",
    "Nathalie", "Sylvie", "Christine", "Véronique", "Maria", "Ana", "Paula",
    "Fernanda", "Giulia", "Chiara", "Elena", "Laura", "Sarah", "Emma",
    "Léa", "Chloé", "Camille", "Charlotte", "Julie", "Céline", "Aurélie",
    "Anna", "Kateryna", "Olena", "Fatima", "Aïcha", "Mariam",
]

# Common last names
LAST_NAMES = [
    "Müller", "Weber", "Schmidt", "Schmit", "Wagner", "Hoffmann", "Klein",
    "Becker", "Meyer", "Kieffer", "Schroeder", "Reding", "Thill", "Reuter",
    "Fernandes", "Silva", "Santos", "Rodrigues", "Pereira", "Costa",
    "Martin", "Bernard", "Dubois", "Thomas", "Robert", "Petit", "Durand",
    "Rossi", "Russo", "Ferrari", "Esposito", "Bianchi",
    "Janssen", "Peeters", "Claes", "Maes",
    "Kowalski", "Nowak", "Popa", "Ionescu",
]

# ROME job codes (simplified list of common job categories)
ROME_CODES = {
    "M1607": ("Secrétariat", ["Secrétaire", "Secrétaire administratif", "Secrétaire juridique"]),
    "N1103": ("Magasinage et préparation de commandes", ["Préparateur de commandes", "Magasinier", "Agent logistique"]),
    "K2204": ("Nettoyage de locaux", ["Agent de nettoyage", "Agent d'entretien", "Technicien de surface"]),
    "D1106": ("Vente en alimentation", ["Vendeur", "Employé de libre-service", "Caissier"]),
    "F1704": ("Préparation du gros œuvre", ["Maçon", "Coffreur", "Ferrailleur"]),
    "H3302": ("Opérations manuelles d'assemblage", ["Agent de fabrication", "Opérateur de production"]),
    "M1602": ("Opérations administratives", ["Agent administratif", "Employé administratif"]),
    "K1304": ("Services domestiques", ["Aide à domicile", "Auxiliaire de vie"]),
    "G1603": ("Personnel polyvalent en restauration", ["Serveur", "Commis de cuisine", "Plongeur"]),
    "I1604": ("Mécanique automobile", ["Mécanicien auto", "Technicien automobile"]),
    "H2502": ("Management et ingénierie de production", ["Ingénieur production", "Responsable production"]),
    "M1805": ("Études et développement informatique", ["Développeur", "Ingénieur logiciel", "Analyste programmeur"]),
    "K1802": ("Développement local", ["Chargé de projet", "Coordinateur"]),
    "C1501": ("Gérance immobilière", ["Gestionnaire locatif", "Agent immobilier"]),
    "E1101": ("Animation de site multimédia", ["Community manager", "Webmaster"]),
}

# Education specialties by ISCED level
EDUCATION_SPECIALTIES = {
    3: ["Commerce", "Comptabilité", "Mécanique", "Électricité", "Hôtellerie", "Administration"],
    4: ["Artisanat", "Éducation", "Technique"],
    5: ["Informatique", "Commerce international", "Gestion", "Marketing"],
    6: ["Informatique", "Économie", "Droit", "Sciences", "Ingénierie", "Communication"],
    7: ["Finance", "Management", "Data Science", "Droit des affaires", "Ingénierie"],
    8: ["Recherche", "Sciences", "Économie appliquée"],
}

DIPLOMA_BY_ISCED = {
    1: "Primaire",
    2: "Secondaire (1er cycle)",
    3: ["DAP", "DT", "Secondaire classique", "Secondaire technique"],
    4: ["Brevet de maîtrise", "Diplôme d'éducateur"],
    5: "BTS",
    6: "Bachelor",
    7: "Master",
    8: "PhD",
}

# Street names for addresses
STREET_NAMES = [
    "rue de la Gare", "avenue de la Liberté", "rue du Commerce", "place Guillaume",
    "rue de Strasbourg", "boulevard Royal", "rue de l'Alzette", "rue de Hollerich",
    "avenue Monterey", "rue des Capucins", "rue Philippe II", "avenue J.F. Kennedy",
    "rue de Bonnevoie", "rue de Beggen", "avenue Pasteur", "rue de Luxembourg",
    "Grand-Rue", "rue Principale", "rue de l'Église", "rue du Parc",
]


# ============================================================================
# HELPER FUNCTIONS
# ============================================================================


def generate_matricule() -> str:
    """Generate a 13-digit Luxembourg matricule."""
    # Format: YYYYMMDDXXXCC (birth date + sequence + check)
    return "".join([str(random.randint(0, 9)) for _ in range(13)])


def generate_birth_date(min_age: int = 18, max_age: int = 65) -> tuple[date, int]:
    """Generate a birth date and return (date, age)."""
    today = date.today()
    age = random.randint(min_age, max_age)
    # Random day in birth year
    birth_year = today.year - age
    birth_month = random.randint(1, 12)
    birth_day = random.randint(1, 28)  # Safe day
    birth_date = date(birth_year, birth_month, birth_day)
    return birth_date, age


def weighted_choice(choices_dict: dict) -> str:
    """Make a weighted random choice from a dictionary."""
    items = list(choices_dict.keys())
    weights = list(choices_dict.values())
    return random.choices(items, weights=weights, k=1)[0]


def get_marital_status_by_age(age: int) -> str:
    """Get marital status correlated with age."""
    if age < 25:
        return random.choices(
            ["Célibataire", "Marié", "PA", "Inconnu"], @
            weights=[0.85, 0.10, 0.03, 0.02],
        )[0]
    elif age < 35:
        return random.choices(
            ["Célibataire", "Marié", "PA", "Divorcé", "Inconnu"],
            weights=[0.45, 0.40, 0.08, 0.05, 0.02],
        )[0]
    elif age < 50:
        return random.choices(
            ["Célibataire", "Marié", "Divorcé", "Séparé", "Veuf", "PA", "Inconnu"],
            weights=[0.15, 0.50, 0.18, 0.08, 0.02, 0.05, 0.02],
        )[0]
    else:
        return random.choices(
            ["Célibataire", "Marié", "Divorcé", "Veuf", "Séparé", "Inconnu"],
            weights=[0.10, 0.45, 0.20, 0.15, 0.08, 0.02],
        )[0]


def get_children_by_age_and_status(age: int, marital_status: str) -> str:
    """Get children indicator correlated with age and marital status."""
    base_prob = 0.0
    if marital_status in ["Marié", "Divorcé", "Veuf", "Séparé"]:
        base_prob = 0.7
    elif marital_status == "PA":
        base_prob = 0.4
    else:
        base_prob = 0.15

    # Adjust by age
    if age < 25:
        base_prob *= 0.3
    elif age < 30:
        base_prob *= 0.6
    elif age > 45:
        base_prob *= 0.9

    if random.random() < base_prob:
        return "Oui"
    elif random.random() < 0.95:
        return "Non"
    else:
        return "Inconnu"


def get_education_level_by_age(age: int) -> int:
    """Get ISCED education level correlated with age (younger = more educated on average)."""
    if age < 30:
        return random.choices(
            [2, 3, 4, 5, 6, 7, 8],
            weights=[0.05, 0.25, 0.10, 0.15, 0.25, 0.18, 0.02],
        )[0]
    elif age < 45:
        return random.choices(
            [1, 2, 3, 4, 5, 6, 7, 8],
            weights=[0.02, 0.10, 0.35, 0.12, 0.12, 0.18, 0.10, 0.01],
        )[0]
    else:
        return random.choices(
            [1, 2, 3, 4, 5, 6, 7],
            weights=[0.05, 0.20, 0.40, 0.10, 0.10, 0.10, 0.05],
        )[0]


def get_language_levels_by_nationality(nationality: str) -> dict[str, str | None]:
    """Get language levels correlated with nationality."""
    levels = ["A", "B", "C"]

    # Base probabilities
    lang = {"FR": None, "LB": None, "DE": None, "EN": None}

    if nationality == "Luxembourgeoise":
        lang["LB"] = random.choices(levels, weights=[0.05, 0.15, 0.80])[0]
        lang["FR"] = random.choices(levels, weights=[0.05, 0.25, 0.70])[0]
        lang["DE"] = random.choices(levels, weights=[0.10, 0.30, 0.60])[0]
        lang["EN"] = random.choices(levels + [None], weights=[0.15, 0.35, 0.30, 0.20])[0]
    elif nationality in ["Française", "Belge"]:
        lang["FR"] = "C"
        lang["LB"] = random.choices(levels + [None], weights=[0.30, 0.20, 0.10, 0.40])[0]
        lang["DE"] = random.choices(levels + [None], weights=[0.25, 0.20, 0.10, 0.45])[0]
        lang["EN"] = random.choices(levels + [None], weights=[0.20, 0.30, 0.20, 0.30])[0]
    elif nationality == "Allemande":
        lang["DE"] = "C"
        lang["FR"] = random.choices(levels + [None], weights=[0.25, 0.25, 0.15, 0.35])[0]
        lang["LB"] = random.choices(levels + [None], weights=[0.35, 0.20, 0.10, 0.35])[0]
        lang["EN"] = random.choices(levels + [None], weights=[0.15, 0.35, 0.30, 0.20])[0]
    elif nationality == "Portugaise":
        lang["FR"] = random.choices(levels, weights=[0.15, 0.40, 0.45])[0]
        lang["LB"] = random.choices(levels + [None], weights=[0.30, 0.25, 0.15, 0.30])[0]
        lang["DE"] = random.choices(levels + [None], weights=[0.35, 0.20, 0.05, 0.40])[0]
        lang["EN"] = random.choices(levels + [None], weights=[0.30, 0.20, 0.10, 0.40])[0]
    elif nationality in ["Britannique"]:
        lang["EN"] = "C"
        lang["FR"] = random.choices(levels + [None], weights=[0.25, 0.30, 0.15, 0.30])[0]
        lang["DE"] = random.choices(levels + [None], weights=[0.35, 0.15, 0.05, 0.45])[0]
        lang["LB"] = random.choices(levels + [None], weights=[0.40, 0.15, 0.05, 0.40])[0]
    else:
        # Other nationalities
        lang["FR"] = random.choices(levels + [None], weights=[0.30, 0.30, 0.20, 0.20])[0]
        lang["EN"] = random.choices(levels + [None], weights=[0.25, 0.30, 0.20, 0.25])[0]
        lang["DE"] = random.choices(levels + [None], weights=[0.40, 0.20, 0.10, 0.30])[0]
        lang["LB"] = random.choices(levels + [None], weights=[0.50, 0.20, 0.05, 0.25])[0]

    return lang


def get_mobility_by_age_and_residence(age: int, is_resident: bool) -> tuple[str, list[str]]:
    """Get mobility and driving permits correlated with age."""
    # Younger people less likely to have car
    if age < 25:
        mobility_weights = [0.15, 0.45, 0.30, 0.10]
    elif age < 40:
        mobility_weights = [0.50, 0.30, 0.15, 0.05]
    else:
        mobility_weights = [0.55, 0.25, 0.12, 0.08]

    mobility_options = [
        "Excellente (permis A, B et véhicule personnel)",
        "Moyenne (permis B mais pas de voiture disponible ou utilisation de transports en communs)",
        "Limitée (pas de voiture et pas de transports en commun disponible)",
        "Limitée pour des raisons psychiques et/ou physiques (durée de trajet restreinte)",
    ]

    mobility = random.choices(mobility_options, weights=mobility_weights)[0]

    # Permits based on mobility and random chance
    permits = []
    if "Excellente" in mobility or "Moyenne" in mobility:
        permits.append("B")
        if random.random() < 0.15:
            permits.append("A")
        if random.random() < 0.08:
            permits.append("C")
        if random.random() < 0.05:
            permits.append("D")

    return mobility, permits


def get_rome_jobs_by_education(isced_level: int, age: int) -> list[dict]:
    """Get ROME job codes correlated with education level."""
    jobs = []

    # Higher education -> more qualified jobs
    if isced_level >= 6:
        eligible_codes = ["M1805", "H2502", "K1802", "C1501", "E1101", "M1607", "M1602"]
    elif isced_level >= 4:
        eligible_codes = ["M1607", "M1602", "I1604", "D1106", "K1304", "G1603"]
    else:
        eligible_codes = ["N1103", "K2204", "D1106", "F1704", "H3302", "K1304", "G1603"]

    # Pick 1-3 jobs
    num_jobs = random.choices([1, 2, 3], weights=[0.40, 0.40, 0.20])[0]
    selected_codes = random.sample(eligible_codes, min(num_jobs, len(eligible_codes)))

    for i, code in enumerate(selected_codes, 1):
        category, appellations = ROME_CODES[code]
        appellation = random.choice(appellations)

        # Experience correlated with age
        max_exp = max(0, (age - 20) * 12)  # months
        exp_months = random.randint(0, max_exp) if max_exp > 0 else 0

        if exp_months >= 24:
            exp_str = f"{exp_months // 12} ans"
        elif exp_months > 0:
            exp_str = f"{exp_months} mois"
        else:
            exp_str = None

        jobs.append({
            "code": f"{code} - {category}",
            "appellation": appellation,
            "experience": exp_str,
        })

    return jobs


def requires_work_authorization(nationality: str) -> bool:
    """Check if nationality requires work authorization in Luxembourg."""
    eu_nationalities = [
        "Luxembourgeoise", "Française", "Allemande", "Belge", "Italienne",
        "Espagnole", "Portugaise", "Polonaise", "Roumaine", "Britannique",
    ]
    return nationality not in eu_nationalities


# ============================================================================
# MAIN GENERATOR
# ============================================================================


def generate_citizen() -> dict[str, Any]:
    """Generate a single citizen record respecting all metadata rules."""
    record = {}

    # Basic identification
    record["Matricule"] = generate_matricule()

    # Gender
    genre = random.choices(["Masculin", "Féminin"], weights=[0.52, 0.48])[0]
    record["Genre"] = genre

    # Names based on gender
    if genre == "Masculin":
        record["Prenom"] = random.choice(FIRST_NAMES_M)
    else:
        record["Prenom"] = random.choice(FIRST_NAMES_F)
    record["Nom"] = random.choice(LAST_NAMES)

    # Birth date and age
    birth_date, age = generate_birth_date()
    record["Date_Naissance"] = birth_date.strftime("%Y-%m-%d")

    # Nationality
    nationality = weighted_choice(NATIONALITIES)
    record["Nationalite"] = nationality

    # Residence type (correlated with nationality)
    if nationality in ["Luxembourgeoise"]:
        residence_type = "résident"
    elif nationality in ["Française", "Belge", "Allemande"]:
        residence_type = random.choices(
            ["résident", "non-résident"],
            weights=[0.40, 0.60],
        )[0]
    else:
        residence_type = random.choices(
            ["résident", "non-résident", "inconnu"],
            weights=[0.75, 0.15, 0.10],
        )[0]
    record["Type_Residence"] = residence_type

    # Address based on residence type
    if residence_type == "résident" or residence_type == "inconnu":
        commune, postal = random.choice(COMMUNES_LU)
        record["Residence_Pays"] = "Luxembourg"
    else:
        # Cross-border worker
        country = random.choice(list(CROSS_BORDER.keys()))
        commune, postal = random.choice(CROSS_BORDER[country])
        record["Residence_Pays"] = country

    record["Residence_Commune"] = commune
    record["Residence_Code_Postal"] = postal
    record["Residence_Localite"] = f"{postal} - {commune}"
    record["Residence_Rue"] = random.choice(STREET_NAMES)
    record["Residence_Num_Rue"] = str(random.randint(1, 200))

    # Birth location (often same country as nationality)
    if nationality == "Luxembourgeoise":
        record["Pays_Naissance"] = "Luxembourg"
        record["Lieu_Naissance"] = random.choice(COMMUNES_LU)[0]
    else:
        country_map = {
            "Française": "France", "Portugaise": "Portugal", "Allemande": "Allemagne",
            "Belge": "Belgique", "Italienne": "Italie", "Espagnole": "Espagne",
            "Britannique": "Royaume-Uni", "Polonaise": "Pologne", "Roumaine": "Roumanie",
            "Cap-Verdienne": "Cap-Vert", "Marocaine": "Maroc",
            "Ukrainienne": "Ukraine", "Syrienne": "Syrie", "Érythréenne": "Érythrée",
        }
        record["Pays_Naissance"] = country_map.get(nationality, "Autre")
        record["Lieu_Naissance"] = ""

    # Marital status and children (correlated)
    marital_status = get_marital_status_by_age(age)
    record["Etat_Civil"] = marital_status
    record["Enfants"] = get_children_by_age_and_status(age, marital_status)

    # Handicap status (small percentage)
    record["Statut_SH"] = random.random() < 0.03
    if record["Statut_SH"]:
        record["Orientation_SH"] = random.choices(
            ["Marché ordinaire", "Atelier protégé", "Inconnu"],
            weights=[0.60, 0.30, 0.10],
        )[0]
    else:
        record["Orientation_SH"] = None

    # Reduced work capacity
    record["Statut_CTR"] = random.random() < 0.05

    # REVIS (social inclusion income) - correlated with employment difficulty
    record["REVIS"] = random.random() < 0.08

    # Work authorization (for non-EU nationals)
    if requires_work_authorization(nationality):
        record["Titre_Sejour"] = random.choice(["065", "100", "109"])
        today = date.today()
        debut = today - timedelta(days=random.randint(30, 365))
        fin = debut + timedelta(days=random.randint(365, 730))
        record["Debut_Titre_Sejour"] = debut.strftime("%Y-%m-%d")
        record["Fin_Titre_Sejour"] = fin.strftime("%Y-%m-%d")
        record["Autorisation_Travail_Debut"] = debut.strftime("%Y-%m-%d")
        record["Autorisation_Travail_Fin"] = fin.strftime("%Y-%m-%d")
        record["Autorisation_Travail_ISCO"] = None
    else:
        record["Titre_Sejour"] = None
        record["Debut_Titre_Sejour"] = None
        record["Fin_Titre_Sejour"] = None
        record["Autorisation_Travail_Debut"] = None
        record["Autorisation_Travail_Fin"] = None
        record["Autorisation_Travail_ISCO"] = None

    # Education (correlated with age)
    isced_level = get_education_level_by_age(age)
    record["Formation_Reussie_ISCED"] = str(isced_level)

    diploma = DIPLOMA_BY_ISCED.get(isced_level, "Inconnu")
    if isinstance(diploma, list):
        diploma = random.choice(diploma)
    record["Formation_Reussie_Diplome"] = diploma

    record["Formation_Reussie_Classe"] = None
    if isced_level <= 2:
        record["Formation_Reussie_Classe"] = random.choice(
            ["5ème primaire", "6ème primaire", "9ème EST", "10ème EST"]
        )

    if isced_level >= 3:
        specialties = EDUCATION_SPECIALTIES.get(isced_level, ["Général"])
        record["Formation_Reussie_Specialite"] = random.choice(specialties)
        record["Formation_Toutes_Specialites"] = record["Formation_Reussie_Specialite"]
    else:
        record["Formation_Reussie_Specialite"] = None
        record["Formation_Toutes_Specialites"] = None

    record["Formation_Reussie_Pays"] = record["Pays_Naissance"]
    record["Formation_Reussie_Preuve"] = random.choices(
        ["Vérifié", "Reconnu", "Homologué", "Déclaré", "A valider"],
        weights=[0.30, 0.25, 0.20, 0.15, 0.10],
    )[0]

    # Languages (correlated with nationality)
    languages = get_language_levels_by_nationality(nationality)
    record["Langue_FR"] = languages["FR"]
    record["Langue_LB"] = languages["LB"]
    record["Langue_DE"] = languages["DE"]
    record["Langue_EN"] = languages["EN"]
    record["Langues_Autres"] = random.choice([None, "Portugais", "Italien", "Espagnol", "Arabe", "Russe"])

    # Mobility and permits
    is_resident = residence_type == "résident"
    mobility, permits = get_mobility_by_age_and_residence(age, is_resident)
    record["Mobilite"] = mobility
    record["Permis"] = ", ".join(permits) if permits else None

    # Job targets (ROME codes)
    jobs = get_rome_jobs_by_education(isced_level, age)
    for i in range(1, 4):
        if i <= len(jobs):
            job = jobs[i - 1]
            record[f"Metier_ROME_{i}"] = job["code"]
            record[f"Metier_ROME_{i}_Appellation"] = job["appellation"]
            record[f"Metier_ROME_{i}_Experience"] = job["experience"]
        else:
            record[f"Metier_ROME_{i}"] = None
            record[f"Metier_ROME_{i}_Appellation"] = None
            record[f"Metier_ROME_{i}_Experience"] = None

    # Work preferences
    record["Temps_De_Travail"] = random.choices(
        ["Temps complet", "Temps partiel", "Sans préférence"],
        weights=[0.65, 0.25, 0.10],
    )[0]

    record["Horaire_Journalier"] = random.choices(
        ["Matin et Après-midi", "Flexible", "Matin", "Après-midi", "Soir", "Après-midi et Soir"],
        weights=[0.40, 0.30, 0.10, 0.10, 0.05, 0.05],
    )[0]

    # Weekly hours (correlated with time preference)
    if record["Temps_De_Travail"] == "Temps complet":
        record["Nbr_Heure_Semaine"] = random.randint(35, 40)
    elif record["Temps_De_Travail"] == "Temps partiel":
        record["Nbr_Heure_Semaine"] = random.randint(15, 30)
    else:
        record["Nbr_Heure_Semaine"] = random.randint(20, 40)

    return record


def generate_dataset(n_citizens: int = 1000) -> pd.DataFrame:
    """Generate a dataset of n citizens."""
    records = [generate_citizen() for _ in range(n_citizens)]
    return pd.DataFrame(records)


if __name__ == "__main__":
    print("Generating synthetic citizen dataset...")

    # Generate 1000 citizens
    df = generate_dataset(1000)

    # Display summary
    print(f"\nGenerated {len(df)} citizens")
    print("\nColumn summary:")
    print(f"  - Total columns: {len(df.columns)}")
    print(f"  - Gender distribution: {df['Genre'].value_counts().to_dict()}")
    print(f"  - Residence type: {df['Type_Residence'].value_counts().to_dict()}")
    print(f"  - Top nationalities: {df['Nationalite'].value_counts().head(5).to_dict()}")
    print(f"  - Education levels (ISCED): {df['Formation_Reussie_ISCED'].value_counts().sort_index().to_dict()}")

    # Save to CSV
    output_file = DATA_PROCESSED / "citizens_synthetic.csv"
    output_file.parent.mkdir(parents=True, exist_ok=True)
    df.to_csv(output_file, index=False)
    print(f"\nDataset saved to: {output_file}")

    # Also display first few rows
    print("\nFirst 5 rows (selected columns):")
    display_cols = ["Matricule", "Nom", "Prenom", "Genre", "Nationalite", "Type_Residence", "Formation_Reussie_ISCED"]
    print(df[display_cols].head().to_string(index=False))



import numpy as np
import pandas as pd


class DatetimeProcessor:
    """Process datetime for avatarization.

    Examples
    --------
    >>> df = pd.DataFrame(
    ...    {
    ...        "date_1": ["2015-01-01 07:00:00", "2015-01-01 10:00:00"],
    ...        "date_2": ["2018-01-01 11:00:00", "2020-01-01 11:00:00"],
    ...    }
    ... )
    >>> df = df.apply(pd.to_datetime, format='%Y-%m-%d %H:%M:%S.%f')
    >>> processor = DatetimeProcessor()
    >>> processor.preprocess(df)
             date_1        date_2
    0  1.420096e+09  1.514804e+09
    1  1.420106e+09  1.577876e+09
    >>> avatar = pd.DataFrame(
    ...    {
    ...        "date_1": [1.420096e+09, 1.537804e+09],
    ...        "date_2": [1.420106e+09, 1.568676e+09],
    ...    }
    ... )
    >>> avatar
             date_1        date_2
    0  1.420096e+09  1.420106e+09
    1  1.537804e+09  1.568676e+09
    >>> processor.postprocess(df, avatar)
                   date_1              date_2
    0 2015-01-01 07:06:40 2015-01-01 09:53:20
    1 2018-09-24 15:46:40 2019-09-16 23:20:00
    """

    def preprocess(self, df: pd.DataFrame) -> pd.DataFrame:
        """Turn datetime variables into int64 (seconds since epoch)."""
        df = df.copy()
        self.datetime_variables = [
            col for col in df.columns if str(df[col].dtype).startswith("datetime")
        ]

        for col in df.columns:
            if col not in self.datetime_variables:
                continue
            epoch = np.datetime64(0, "Y")  # 0 years since epoch ==> epoch itself
            df[col] = (df[col].to_numpy() - epoch) / np.timedelta64(10**9, "ns")
            df[col] = df[col].astype(float)  # Necessary, else its float64
        return df

    def postprocess(self, source: pd.DataFrame, dest: pd.DataFrame) -> pd.DataFrame:
        """Transform datetime columns from seconds since epoch back to datetime[ns] type."""
        default_value = 0

        for name in self.datetime_variables:
            # Replace NaN values with default_value, convert to datetime[ns] and replace with NaT
            # afterwards
            is_nan = dest[name].isna()
            dest[name] = dest[name].fillna(default_value)
            dest[name] = pd.to_datetime(dest[name].astype("int64") * 10**9, unit="ns")
            dest[name] = dest[name].where(~is_nan, pd.NaT)
        return dest



import os
import time
import warnings
import webbrowser
from dataclasses import asdict
from logging import getLogger
from pathlib import Path
from typing import Any, Optional, Union

import pandas as pd
import tenacity
from avatar_yaml import (
    AvatarizationFastDPParameters,
    AvatarizationOpenDPParameters,
    AvatarizationParameters,
    PrivacyMetricsParameters,
)
from avatar_yaml import Config as Config
from avatar_yaml.models.advice import AdviceType
from avatar_yaml.models.avatar_metadata import (
    DataRecipient,
    DataSubject,
    DataType,
    SensitivityLevel,
)
from avatar_yaml.models.parameters import (
    AlignmentMethod,
    AugmentationStrategy,
    ExcludeVariablesMethod,
    ImputeMethod,
    InterRecordRangeDifferenceParameters,
    ProjectionType,
    ReportType,
)
from avatar_yaml.models.schema import ColumnType, LinkMethod
from IPython.display import HTML, display

from avatars import __version__
from avatars.client import ApiClient
from avatars.constants import (
    DEFAULT_DELAY_BETWEEN_CONSECUTIVE_JOBS,
    DEFAULT_POLL_INTERVAL,
    DEFAULT_TYPE,
    ERROR_STATUSES,
    JOB_EXECUTION_ORDER,
    MATCHERS,
    RESULTS_TO_STORE,
    VOLUME_NAME,
    PlotKind,
    Results,
    TypeResults,
)
from avatars.crash_handler import register_runner
from avatars.file_downloader import FileDownloader
from avatars.job_launcher import JobLauncher
from avatars.models import BulkDeleteRequest, BulkDeleteResponse, JobKind, JobResponse
from avatars.results_organizer import ResultsOrganizer

logger = getLogger(__name__)


class Runner:
    def __init__(
        self,
        api_client: ApiClient,
        display_name: str,
        seed: int | None = None,
        max_distribution_plots: int | None = None,
        pia_data_recipient: DataRecipient = DataRecipient.UNKNOWN,
        pia_data_type: DataType = DataType.UNKNOWN,
        pia_data_subject: DataSubject = DataSubject.UNKNOWN,
        pia_sensitivity_level: SensitivityLevel = SensitivityLevel.UNDEFINED,
    ) -> None:
        self.client = api_client
        self.display_name = display_name
        self.set_name: str | None = None
        self.config = Config(
            set_name=self.display_name, seed=seed, max_distribution_plots=max_distribution_plots
        )
        self.file_downloader = FileDownloader(api_client)
        self.results: ResultsOrganizer = ResultsOrganizer()
        self.jobs = JobLauncher(api_client, self.config)
        self.results_urls: dict[str, dict[str, list[str]]] = {}

        annotations = {
            "client_type": "python",
            "client_version": __version__,
        }
        self.config.create_metadata(
            annotations,
            pia_datarecipient=pia_data_recipient,
            pia_datatype=pia_data_type,
            pia_datasubject=pia_data_subject,
            pia_sensitivitylevel=pia_sensitivity_level,
        )

        # Register this Runner instance for crash reporting
        register_runner(self)

    def add_annotations(self, annotations: dict[str, str]) -> None:
        """Add metadata annotations to the config.

        Parameters
        ----------
        annotations
            A dictionary of annotations to add to the metadata.
        """
        if self.config.avatar_metadata is None:
            self.config.create_metadata(annotations)
        else:
            current_annotations = self.config.avatar_metadata.annotations or {}
            current_annotations.update(annotations)
            self.config.create_metadata(current_annotations)

    def add_table(
        self,
        table_name: str,
        data: str | pd.DataFrame,
        primary_key: str | None = None,
        foreign_keys: list | None = None,
        time_series_time: str | None = None,
        types: dict[str, ColumnType] = {},
        individual_level: bool | None = None,
        avatar_data: str | pd.DataFrame | None = None,
    ):
        """Add a table to the config and upload the data in the server.

        Parameters
        ----------
        table_name
            The name of the table.
        data
            The data to add to the table. Can be a path to a file or a pandas DataFrame.
        primary_key
            The primary key of the table.
        foreign_keys
            Foreign keys of the table.
        time_series_time
            name of the time column in the table (time series case).
        types
            A dictionary of column types with the column name as the key and the type as the value.
        individual_level
            A boolean as true if the table is at individual level or not. An individual level table
            is a table where each row corresponds to an individual (ex: patient, customer, etc.).
            Default behavior is True.
        avatar_data
            The avatar table if there is one. Can be a path to a file or a pandas DataFrame.
        """
        file, avatar_file = self.upload_file(table_name, data, avatar_data)
        if isinstance(data, pd.DataFrame):
            types = self._get_types(data, types)

        if foreign_keys == [None]:
            foreign_keys = None

        self.config.create_table(
            table_name=table_name,
            original_volume=VOLUME_NAME,
            original_file=file,
            avatar_volume=VOLUME_NAME if avatar_data is not None else None,
            avatar_file=avatar_file,
            primary_key=primary_key,
            foreign_keys=foreign_keys,
            time_series_time=time_series_time,
            types=types,
            individual_level=individual_level,
        )

    def advise_parameters(self, table_name: str | None = None) -> None:
        """Fill the parameters set with the server recommendation.

        Parameters
        ----------
        table_name
            The name of the table. If None, all tables will be used.
        """
        self._setup_advice_config()
        if table_name:
            tables = [table_name]
        else:
            tables = list(self.config.tables.keys())
        self._create_advice_jobs(tables)
        self._apply_advice_parameters(tables)

    def _setup_advice_config(self) -> None:
        """Create advice config and upload resources to server."""
        self.config.create_advice(advisor_type=[AdviceType.PARAMETERS])
        yaml = self.config.get_yaml()
        resource_response = self.client.resources.put_resources(
            display_name=self.display_name,
            yaml_string=yaml,
        )
        # Update set_name with the actual UUID returned by the backend
        self.set_name = str(resource_response.set_name)

    def _create_advice_jobs(self, tables: list[str]) -> None:
        """Download advice results for the specified tables."""
        for table_name in tables:
            if self.results.advice.get(table_name) is None:
                # Only launch the advice job once, not per table
                # but relaunch it if a table didn't get advice yet
                if self.set_name is None:
                    raise ValueError("Set name is not set. Cannot launch advice job.")
                self.jobs.launch_job(JobKind.advice, self.set_name)
                job_name = self.jobs.get_parameters_name(JobKind.advice)
                self._download_specific_result(job_name, Results.ADVICE)

    def _apply_advice_parameters(self, tables: list[str]) -> None:
        """Apply the downloaded advice parameters to each table."""
        job_name = self.jobs.get_parameters_name(JobKind.advice)
        for table_name in tables:
            advise_parameters = self.results.get_results(table_name, Results.ADVICE, job_name)
            if not isinstance(advise_parameters, dict):
                raise ValueError("Expected advice parameters to be a dictionary")

            imputation_data: dict[str, Any] = {}
            if advise_parameters.get("imputation") is not None:
                imputation_result = advise_parameters.get("imputation")
                if isinstance(imputation_result, dict):
                    imputation_data = imputation_result

            imputation_method = None
            if imputation_data.get("method") is not None:
                imputation_method = ImputeMethod(imputation_data.get("method"))

            self.set_parameters(
                k=advise_parameters.get("k"),
                use_categorical_reduction=advise_parameters.get("use_categorical_reduction"),
                ncp=advise_parameters.get("ncp"),
                imputation_method=imputation_method,
                imputation_k=imputation_data.get("k"),
                imputation_training_fraction=imputation_data.get("training_fraction"),
                imputation_return_data_imputed=imputation_data.get("return_data_imputed", False),
                table_name=table_name,
            )

    def upload_file(
        self,
        table_name: str,
        data: str | pd.DataFrame,
        avatar_data: str | pd.DataFrame | None = None,
    ):
        """Upload a file to the server.

        Parameters
        ----------
        data
            The data to upload. Can be a path to a file or a pandas DataFrame.
        file_name
            The name of the file.
        """
        extension = ".csv" if isinstance(data, pd.DataFrame) else Path(data).suffix
        file = table_name + extension
        self.client.upload_file(data=data, key=file)
        avatar_file = None
        if avatar_data is not None:
            avatar_file = table_name + "_avatars" + extension
            self.client.upload_file(data=avatar_data, key=avatar_file)
        return file, avatar_file

    def _get_types(
        self, data: pd.DataFrame, types: dict[str, ColumnType] = {}
    ) -> dict[str, ColumnType]:
        dtypes = {}
        for column_name, _type in data.dtypes.items():
            column_name_str = str(column_name)
            if column_name_str in types:
                dtypes[column_name_str] = types[column_name_str]
            else:
                dtypes[column_name_str] = self._get_type_from_pandas(str(_type))
        return dtypes

    def _get_type_from_pandas(self, value: str) -> ColumnType:
        """Return our data type from pandas type."""
        for matcher, our_type in MATCHERS.items():
            if matcher.search(value):
                return our_type
        return DEFAULT_TYPE

    def add_link(
        self,
        parent_table_name: str,
        parent_field: str,
        child_table_name: str,
        child_field: str,
        method: LinkMethod = LinkMethod.LINEAR_SUM_ASSIGNMENT,
    ):
        """Add a table link to the config.

        Parameters
        ----------
        parent_table_name
            The name of the parent table.
        child_table_name
            The name of the child table.
        parent_field
            The parent link key field (primary key) in the parent table.
        child_field
            The child link key field (foreign key)in the child table.
        method
            The method to use for linking the tables. Defaults to "linear_sum_assignment".
        """
        self.config.create_link(
            parent_table_name, child_table_name, parent_field, child_field, method.value
        )

    def set_parameters(
        self,
        table_name: str,
        k: int | None = None,
        ncp: int | None = None,
        use_categorical_reduction: bool | None = None,
        column_weights: dict[str, float] | None = None,
        exclude_variable_names: list[str] | None = None,
        exclude_replacement_strategy: ExcludeVariablesMethod | None = None,  # DEPRECATED
        exclude_variable_method: ExcludeVariablesMethod | None = None,
        imputation_method: ImputeMethod | None = None,
        imputation_k: int | None = None,
        imputation_training_fraction: float | None = None,
        imputation_return_data_imputed: bool | None = None,
        open_dp_epsilon: float | None = None,
        fast_dp_epsilon: float | None = None,
        time_series_nf: int | None = None,
        time_series_projection_type: ProjectionType | None = None,
        time_series_nb_points: int | None = None,
        time_series_method: AlignmentMethod | None = None,
        known_variables: list[str] | None = None,
        target: str | None = None,
        quantile_threshold: int | None = None,
        data_augmentation_strategy: float | AugmentationStrategy | dict[str, float] | None = None,
        data_augmentation_target_column: str | None = None,
        data_augmentation_should_anonymize_original_table: bool | None = None,
        inter_record_processors: list[InterRecordRangeDifferenceParameters] | None = None,
    ):
        """Set the parameters for a given table.

        This will overwrite any existing parameters for the table, including parameters set using
        `advise_parameter()`.

        Parameters
        ----------
        table_name
            The name of the table.
        k
            Number of nearest neighbors to consider for KNN-based methods.
        ncp
            Number of dimensions to consider for the KNN algorithm.
        use_categorical_reduction
            Whether to transform categorical variables into a
            latent numerical space before projection.
        column_weights
            Dictionary mapping column names to their respective weights,
            indicating the importance of each variable during the projection process.
        exclude_variable_names
            List of variable names to exclude from the projection.
        exclude_replacement_strategy:
            DEPRECATED: use exclude_variable_method instead.
        exclude_variable_method:
            Strategy for replacing excluded variables. Options: ROW_ORDER, COORDINATE_SIMILARITY.
        imputation_method
            Method for imputing missing values. Options: ``ImputeMethod.KNN``,
            ``ImputeMethod.MODE``, ``ImputeMethod.MEDIAN``, ``ImputeMethod.MEAN``,
            ``ImputeMethod.FAST_KNN``.
        imputation_k
            Number of neighbors to use for imputation if the method is KNN or FAST_KNN.
        imputation_training_fraction
            Fraction of the dataset to use for training the imputation model
            when using KNN or FAST_KNN.
        imputation_return_data_imputed:
            Whether to return the data with imputed values.
        open_dp_epsilon
            Epsilon value for differential privacy using OpenDP implementation.
        fast_dp_epsilon
            Epsilon value for fastDP avatarization.
        time_series_nf
            In time series context, number of degrees of freedom to
            retain in time series projections.
        time_series_projection_type
            In time series context, type of projection for time series. Options:
            ``ProjectionType.FCPA`` (default) or ``ProjectionType.FLATTEN``.
        time_series_method
            In time series context, method for aligning series. Options:
            ``AlignmentMethod.SPECIFIED``, ``AlignmentMethod.MAX``,
            ``AlignmentMethod.MIN``, ``AlignmentMethod.MEAN``.
        time_series_nb_points
            In time series context, number of points to generate for time series.
        known_variables
            List of known variables to be used for privacy metrics.
            These are variables that could be easily known by an attacker.
        target
            Target variable to predict, used for signal metrics.
        quantile_threshold
            Quantile threshold for privacy metrics calculations.
        data_augmentation_strategy
            Strategy for data augmentation. Can be a float representing the
            augmentation ratio, an AugmentationStrategy enum, or a dictionary
            mapping modality to their respective augmentation ratios.
        data_augmentation_target_column
            Target column for data augmentation when using a dictionary strategy or
            AugmentationStrategy.
        data_augmentation_should_anonymize_original_table
            SENSITIVE: Whether to anonymize the original table during data augmentation.
            Default is True.
        inter_record_processors
            List of InterRecordRangeDifferenceParameters objects. Each processor transforms
            a pair of start/end columns.

        The processor transforms start/end column pairs into internal representation.
        This can lead to better semantic avatarization.The transformation
        is transparent to the user - input and output have the same column structure at the end.

        Records are automatically sorted by target_start_variable for processing.
        """
        imputation = imputation_method.value if imputation_method else None
        if exclude_variable_method:
            replacement_strategy = exclude_variable_method.value
        elif exclude_replacement_strategy:
            warnings.warn(
                "The 'exclude_replacement_strategy' parameter is deprecated and will be removed "
                "in a future release. Please use 'exclude_variable_method' instead."
            )
            replacement_strategy = exclude_replacement_strategy.value
        else:
            replacement_strategy = None
        if k and open_dp_epsilon:
            raise ValueError(
                "Expected either k or open_dp_epsilon to be set, not both. "
                "If you want to use OpenDP avatarization, set open_dp_epsilon and remove k."
            )
        if k and fast_dp_epsilon:
            raise ValueError(
                "Expected either k or fast_dp_epsilon to be set, not both. "
                "If you want to use fastDP avatarization, set fast_dp_epsilon and remove k."
            )
        if open_dp_epsilon and fast_dp_epsilon:
            raise ValueError(
                "Expected either open_dp_epsilon or fast_dp_epsilon to be set, not both. "
                "Choose either OpenDP (open_dp_epsilon) or FastDP (fast_dp_epsilon)."
            )

        inter_record_processor_params_list = inter_record_processors

        # reset the parameters if they were already set
        if self.config.avatarization and self.config.avatarization.get(table_name):
            del self.config.avatarization[table_name]
        avatarization_open_dp = getattr(self.config, "avatarization_open_dp", None)
        if avatarization_open_dp and avatarization_open_dp.get(table_name):
            del avatarization_open_dp[table_name]
        avatarization_fast_dp = getattr(self.config, "avatarization_fast_dp", None)
        if avatarization_fast_dp and avatarization_fast_dp.get(table_name):
            del avatarization_fast_dp[table_name]
        if self.config.privacy_metrics and self.config.privacy_metrics.get(table_name):
            del self.config.privacy_metrics[table_name]
        if self.config.signal_metrics and self.config.signal_metrics.get(table_name):
            del self.config.signal_metrics[table_name]
        if self.config.time_series and self.config.time_series.get(table_name):
            del self.config.time_series[table_name]
        if k:
            # Avatarizaztion with avatar method
            self.config.create_avatarization_parameters(
                table_name=table_name,
                k=k,
                ncp=ncp,
                use_categorical_reduction=use_categorical_reduction,
                imputation_method=imputation,
                imputation_k=imputation_k,
                imputation_training_fraction=imputation_training_fraction,
                imputation_return_data_imputed=imputation_return_data_imputed,
                column_weights=column_weights,
                exclude_variable_names=exclude_variable_names,
                exclude_variable_method=replacement_strategy,
                data_augmentation_strategy=data_augmentation_strategy,
                data_augmentation_target_column=data_augmentation_target_column,
                data_augmentation_should_anonymize_original_table=data_augmentation_should_anonymize_original_table,
                avatarization_processors_parameters=inter_record_processor_params_list,
            )

        elif open_dp_epsilon:
            # OpenDP avatarization (MST method)
            self.config.create_avatarization_open_dp_parameters(
                table_name=table_name,
                epsilon=open_dp_epsilon,
                ncp=ncp,
                use_categorical_reduction=use_categorical_reduction,
                imputation_method=imputation,
                imputation_k=imputation_k,
                imputation_training_fraction=imputation_training_fraction,
                imputation_return_data_imputed=imputation_return_data_imputed,
                column_weights=column_weights,
                exclude_variable_names=exclude_variable_names,
                exclude_variable_method=replacement_strategy,
                data_augmentation_strategy=data_augmentation_strategy,
                data_augmentation_target_column=data_augmentation_target_column,
                data_augmentation_should_anonymize_original_table=data_augmentation_should_anonymize_original_table,
            )

        elif fast_dp_epsilon:
            # FastDP avatarization (DP mechanisms in latent space)
            self.config.create_avatarization_fast_dp_parameters(
                table_name=table_name,
                epsilon=fast_dp_epsilon,
                ncp=ncp,
                use_categorical_reduction=use_categorical_reduction,
                imputation_method=imputation,
                imputation_k=imputation_k,
                imputation_training_fraction=imputation_training_fraction,
                imputation_return_data_imputed=imputation_return_data_imputed,
                column_weights=column_weights,
                exclude_variable_names=exclude_variable_names,
                exclude_variable_method=replacement_strategy,
                data_augmentation_strategy=data_augmentation_strategy,
                data_augmentation_target_column=data_augmentation_target_column,
                data_augmentation_should_anonymize_original_table=data_augmentation_should_anonymize_original_table,
                avatarization_processors_parameters=inter_record_processor_params_list,
            )

        if (
            time_series_nf
            or time_series_projection_type
            or time_series_nb_points
            or time_series_method
        ):
            method = time_series_method.value if time_series_method else None
            projection_type = (
                time_series_projection_type.value if time_series_projection_type else None
            )
            self.config.create_time_series_parameters(
                table_name=table_name,
                nf=time_series_nf,
                projection_type=projection_type,
                nb_points=time_series_nb_points,
                method=method,
            )

        self.config.create_privacy_metrics_parameters(
            table_name=table_name,
            ncp=ncp,
            use_categorical_reduction=use_categorical_reduction,
            imputation_method=imputation,
            imputation_k=imputation_k,
            imputation_training_fraction=imputation_training_fraction,
            imputation_return_data_imputed=imputation_return_data_imputed,
            exclude_variable_names=exclude_variable_names,
            exclude_variable_method=replacement_strategy,
            known_variables=known_variables,
            target=target,
            quantile_threshold=quantile_threshold,
            column_weights=column_weights,
        )

        self.config.create_signal_metrics_parameters(
            table_name=table_name,
            ncp=ncp,
            use_categorical_reduction=use_categorical_reduction,
            imputation_method=imputation,
            imputation_k=imputation_k,
            imputation_training_fraction=imputation_training_fraction,
            imputation_return_data_imputed=imputation_return_data_imputed,
            exclude_variable_names=exclude_variable_names,
            exclude_variable_method=replacement_strategy,
            column_weights=column_weights,
        )

    def update_parameters(self, table_name: str, **kwargs) -> None:
        """
        Update specific parameters for the table while preserving other existing parameters.
        Only updates the parameters that are provided, keeping existing values for others.

        Parameters
        ----------
        table_name
            The name of the table.
        **kwargs
            The parameters to update. Only parameters that are provided will be updated.
            See set_parameters for the full list of available parameters.
        """
        avatarization = getattr(self.config, "avatarization", None)
        avatarization_open_dp = getattr(self.config, "avatarization_open_dp", None)
        avatarization_fast_dp = getattr(self.config, "avatarization_fast_dp", None)
        if (
            (avatarization.get(table_name) if avatarization else None) is None
            and (avatarization_open_dp.get(table_name) if avatarization_open_dp else None) is None
            and (avatarization_fast_dp.get(table_name) if avatarization_fast_dp else None) is None
        ):
            raise ValueError(
                f"No existing parameters found for table '{table_name}'. "
                "Use set_parameters to create new parameters."
            )
        # Get current parameters for this table
        current_params = self._extract_current_parameters(table_name)

        # Update only the parameters that were provided
        for param_name, param_value in kwargs.items():
            current_params[param_name] = param_value
        # Apply all parameters back using set_parameters
        self.set_parameters(table_name=table_name, **current_params)

    def _extract_current_parameters(self, table_name: str) -> dict:
        """Extract the current parameters for a given table.

        Parameters
        ----------
        table_name
            The name of the table.

        Returns
        -------
        dict
            A dictionary containing the current parameters for
            the table as it is used in set_parameters.
        """

        current_params: dict[str, Any] = {}

        # Extract avatarization parameters
        if (
            self.config.avatarization is not None
            and table_name in self.config.avatarization.keys()
        ):
            # Standard avatarization parameters
            params: Optional[
                Union[
                    AvatarizationParameters,
                    AvatarizationOpenDPParameters,
                    AvatarizationFastDPParameters,
                    PrivacyMetricsParameters,
                ]
            ] = self.config.avatarization[table_name]
            if isinstance(params, AvatarizationParameters):
                current_params.update(
                    {
                        "k": params.k,
                        "column_weights": params.column_weights,
                        "use_categorical_reduction": params.use_categorical_reduction,
                        "ncp": params.ncp,
                    }
                )
                current_params.update(self._extract_exclude_parameters(params))
        elif (
            hasattr(self.config, "avatarization_open_dp")
            and self.config.avatarization_open_dp is not None
            and table_name in self.config.avatarization_open_dp.keys()
        ):
            # OpenDP avatarization parameters (end-to-end DP)
            params = self.config.avatarization_open_dp[table_name]
            if isinstance(params, AvatarizationOpenDPParameters):
                current_params.update(
                    {
                        "open_dp_epsilon": params.epsilon if params.epsilon else None,
                        "column_weights": params.column_weights,
                        "use_categorical_reduction": params.use_categorical_reduction,
                        "ncp": params.ncp,
                    }
                )
                current_params.update(self._extract_exclude_parameters(params))
        elif (
            hasattr(self.config, "avatarization_fast_dp")
            and self.config.avatarization_fast_dp is not None
            and table_name in self.config.avatarization_fast_dp.keys()
        ):
            # FastDP avatarization parameters
            # Only epsilon is exposed; mechanism-specific params use defaults
            params = self.config.avatarization_fast_dp[table_name]
            if isinstance(params, AvatarizationFastDPParameters):
                current_params.update(
                    {
                        "fast_dp_epsilon": params.epsilon if params.epsilon else None,
                        "column_weights": params.column_weights,
                        "use_categorical_reduction": params.use_categorical_reduction,
                        "ncp": params.ncp,
                    }
                )
                current_params.update(self._extract_exclude_parameters(params))
        else:
            params = None  # No parameters has been preset

        if (
            self.config.privacy_metrics is not None
            and self.config.privacy_metrics.get(table_name) is not None
        ):
            params = self.config.privacy_metrics[table_name]
            current_params.update(
                {
                    "use_categorical_reduction": params.use_categorical_reduction,
                    "ncp": params.ncp,
                }
            )
        else:
            params = None  # No parameters has been preset

        # Extract imputation parameters
        if params and params.imputation:
            current_params.update(
                {
                    "imputation_method": ImputeMethod(params.imputation["method"])
                    if params.imputation["method"]
                    else None,
                    "imputation_k": params.imputation["k"] if params.imputation["k"] else None,
                    "imputation_training_fraction": params.imputation["training_fraction"]
                    if params.imputation["training_fraction"]
                    else None,
                    "imputation_return_data_imputed": params.imputation["return_data_imputed"],
                }
            )

        # Extract time series parameters
        if self.config.time_series and table_name in self.config.time_series.keys():
            ts_params = self.config.time_series[table_name]

            # Projection parameters
            if ts_params.projection:
                current_params.update(
                    {
                        "time_series_nf": ts_params.projection["nf"]
                        if ts_params.projection["nf"]
                        else None,
                        "time_series_projection_type": ProjectionType(
                            ts_params.projection["projection_type"]
                        )
                        if ts_params.projection["projection_type"]
                        else None,
                    }
                )

            # Alignment parameters
            if ts_params.alignment:
                current_params.update(
                    {
                        "time_series_nb_points": ts_params.alignment["nb_points"]
                        if ts_params.alignment["nb_points"]
                        else None,
                        "time_series_method": AlignmentMethod(ts_params.alignment["method"])
                        if ts_params.alignment["method"]
                        else None,
                    }
                )

        # Extract privacy metrics parameters
        if (
            self.config.privacy_metrics is not None
            and table_name in self.config.privacy_metrics.keys()
        ):
            pm_params = self.config.privacy_metrics[table_name]
            to_update = {
                "known_variables": pm_params.known_variables,
                "target": pm_params.target,
                "quantile_threshold": pm_params.quantile_threshold,
            }
            current_params.update(to_update)

        return current_params

    def _extract_exclude_parameters(self, params) -> dict:
        """Extract exclude variables parameters from parameter object.

        Parameters
        ----------
        params:
            The parameters object that contains exclude_variables information.

        Returns
        -------
        A dictionary containing exclude_variable_names and exclude_replacement_strategy parameters.
        """
        result = {}
        if params.exclude_variables:
            result["exclude_variable_names"] = (
                params.exclude_variables["variable_names"]
                if params.exclude_variables["variable_names"]
                else None
            )
            result["exclude_replacement_strategy"] = (
                ExcludeVariablesMethod(params.exclude_variables["replacement_strategy"])
                if params.exclude_variables["replacement_strategy"]
                else None
            )
        return result

    def delete_parameters(self, table_name: str, parameters_names: list[str] | None = None):
        """Delete parameters from the config.

        Parameters
        ----------
        table_name
            The name of the table.
        parameters_names
            The names of the parameters to delete. If None, all parameters will be deleted.
        """
        self.config.delete_parameters(table_name, parameters_names)

    def delete_link(self, parent_table_name: str, child_table_name: str):
        """Delete a link from the config.

        Parameters
        ----------
        parent_table_name
            The name of the parent table.
        child_table_name
            The name of the child table.
        """
        self.config.delete_link(parent_table_name, child_table_name)

    def delete_table(self, table_name: str):
        """Delete a table from the config.

        Parameters
        ----------
        table_name
            The name of the table.
        """
        self.config.delete_table(table_name)

    def get_yaml(self, path: str | None = None):
        """Get the yaml config.

        Parameters
        ----------
        path
            The path to the yaml file. If None, the default config will be returned.
        """
        return self.config.get_yaml(path)

    def _handle_existing_results(self, ignore_warnings: bool = False) -> None:
        """Warn and clean up local state when re-running a runner that already has results."""
        advice_job_name = self.jobs.get_parameters_name(JobKind.advice)
        has_run_before = (
            not self.results.is_results_empty()
            or any(key != advice_job_name for key in self.results_urls)
            or any(job != advice_job_name for job in self.jobs.get_launched_jobs())
        )

        if has_run_before:
            if not ignore_warnings:
                set_name = self.set_name
                warnings.warn(
                    f"\nThis runner already has results for set_name '{set_name}'.\n"
                    "Re-running will create a new set_name — previous results will no longer "
                    "be accessible from this runner.\n"
                    f"To access previous results later:\n"
                    f"regenerated_runner = manager.create_runner_from_id('{set_name}')\n"
                    "regenerated_runner.shuffled(...)",
                    UserWarning,
                )
            self.results_urls.clear()
            self.results = ResultsOrganizer()
            self.jobs.jobs.clear()

    def run(self, jobs_to_run: list[JobKind] = JOB_EXECUTION_ORDER, ignore_warnings: bool = False):
        """Run avatarization jobs.

        This method creates resources and launches the specified jobs in the correct order.
        If this runner has existing results from a previous run, a ``UserWarning`` is emitted
        with the old ``set_name`` so you can recover previous results if needed. Local state
        is then cleared and a new run starts.

        Parameters
        ----------
        jobs_to_run : list[JobKind]
            List of job types to run. Defaults to all jobs in execution order.
        ignore_warnings: bool
            Whether to ignore warnings about existing results when re-running a runner.
            Defaults to False.

        Examples
        --------
        >>> runner = manager.create_runner("my_dataset")
        >>> runner.add_table("customers", data=df)
        >>> runner.run()
        >>>
        >>> # Running again emits a UserWarning with the old set_name, then proceeds
        >>> runner.run()
        """
        self._handle_existing_results(ignore_warnings=ignore_warnings)

        # Create report configurations if report job is requested
        if JobKind.report in jobs_to_run:
            self.config.create_report()
            self.config.create_report(ReportType.PIA)

        yaml = self.get_yaml()

        resource_response = self.client.resources.put_resources(
            display_name=self.display_name,
            yaml_string=yaml,
        )
        # Update set_name with the actual UUID returned by the backend
        self.set_name = str(resource_response.set_name)
        self.jobs.set_name = self.set_name

        # Execute jobs in order
        jobs_to_run = sorted(jobs_to_run, key=lambda job: JOB_EXECUTION_ORDER.index(job))

        for i, job_kind in enumerate(jobs_to_run):
            # Add small delay between job creations to avoid bursts in api calls
            # Skip delay for first job
            if i > 0:
                time.sleep(DEFAULT_DELAY_BETWEEN_CONSECUTIVE_JOBS)
            self.jobs.launch_job(job_kind, self.set_name)

    def get_job(self, job_name: JobKind | str) -> JobResponse:
        """
        Get the job by name.

        Parameters
        ----------
        job_name
            The name of the job to get.
        """
        return self.jobs.get_job_status(job_name)

    def get_status(self, job_name: JobKind):
        """
        Get the status of a job by name.
        Parameters
        ----------
        job_name
            The name of the job to get.
        """
        return self.get_job(job_name).status

    def delete(
        self, job_names: JobKind | str | list[JobKind | str] | None = None
    ) -> BulkDeleteResponse:
        """Delete one or more jobs launched by this runner.

        When called without arguments, every job launched by this runner is deleted.

        Parameters
        ----------
        job_names
            A single job kind/name, a list of job kinds/names, or ``None`` to
            delete all launched jobs.

        Returns
        -------
        BulkDeleteResponse
            Response containing deleted and failed jobs.

        Examples
        --------
        >>> runner.delete()                                              # all jobs
        >>> runner.delete(JobKind.standard)                             # single job
        >>> runner.delete([JobKind.standard, JobKind.privacy_metrics])  # multiple jobs
        """
        if job_names is None:
            names: list[JobKind | str] = list(self.jobs.get_launched_jobs())
        elif not isinstance(job_names, list):
            names = [job_names]
        else:
            names = job_names
        job_ids = [self.jobs.get_job_id(name) for name in names]
        return self.client.jobs.bulk_delete_jobs(BulkDeleteRequest(job_names=job_ids))

    def _retrieve_job_result_urls(self, job_name: str) -> None:
        """
        Get the result of a job by name.

        Parameters
        ----------
        job_name
            The name of the job to get.
        """

        def check_job_status() -> JobResponse:
            """Check job status and raise if in error state, otherwise return if not done."""
            job = self.get_job(job_name)

            if job.status in ERROR_STATUSES:
                if job.exception:
                    raise ValueError(f"Job {job_name} failed with exception: {job.exception}")
                raise ValueError("internal error")

            if not job.done:
                # Raise to trigger retry
                raise tenacity.TryAgain

            return job

        # Use tenacity to poll with exponential backoff capped at 20s
        # Starts at 5s, 20% longer every time, maxes out at 20s
        for attempt in tenacity.Retrying(
            wait=tenacity.wait_exponential(
                min=DEFAULT_POLL_INTERVAL,
                max=4 * DEFAULT_POLL_INTERVAL,
                exp_base=1.2,
            ),
            retry=tenacity.retry_if_exception_type(tenacity.TryAgain),
            reraise=True,
        ):
            with attempt:
                check_job_status()

        job_id = self.jobs.get_job_id(job_name)
        self.results_urls[job_name] = self.client.results.get_results(job_id)

    def _populate_results_from_existing_jobs(self) -> None:
        """Fetch and populate results URLs from all completed jobs for this set_name.

        This method is used when reconstructing a Runner from an existing set_name.
        It fetches all jobs associated with the set_name and populates the results_urls
        dictionary for completed jobs only. It also registers these jobs with the
        JobLauncher so they can be accessed via normal result methods.
        """
        if not self.set_name:
            return

        all_jobs = self.client.jobs.get_jobs()

        # Filter jobs by set_name and only include completed jobs
        for job in all_jobs.jobs:
            if str(job.set_name) == self.set_name and job.done and not job.exception:
                job_name = job.name
                parameters_name = job.parameters_name
                try:
                    self.results_urls[parameters_name] = self.client.results.get_results(job_name)

                    self.jobs.register_existing_job(parameters_name, job_name, f"/jobs/{job_name}")

                except (ValueError, KeyError, RuntimeError) as e:
                    logger.debug(f"Could not fetch results for job {job_name}: {e}")
                    continue

    def get_specific_result_urls(
        self,
        job_name: str,
        result: Results = Results.SHUFFLED,
    ) -> list[str]:
        if not self.jobs.has_job(job_name):
            raise ValueError(f"Expected job '{job_name}' to be created. Try running it first.")
        if job_name not in self.results_urls:
            self._retrieve_job_result_urls(job_name)
        if result not in self.results_urls[job_name]:
            return []
        return self.results_urls[job_name][result]

    def _download_all_files(self):
        for job_name in self.jobs.get_launched_jobs():
            if not self.results_urls or job_name not in self.results_urls.keys():
                self._retrieve_job_result_urls(job_name)
            for result in self.results_urls[job_name].keys():
                for table_name in self.config.tables.keys():
                    if result in Results:
                        if Results(result) in RESULTS_TO_STORE:
                            self.get_specific_result(table_name, job_name, Results(result))

    def _download_specific_result(
        self,
        job_name: str,
        result_name: Results,
    ) -> None:
        urls = self.get_specific_result_urls(job_name=job_name, result=result_name)
        # Use batch download to get all file credentials at once
        downloaded_results = self.file_downloader.download_files_batch(urls)
        for url, result in downloaded_results.items():
            metadata = self._get_metadata(url, result_name, job_name)
            table_name = self.results.get_table_name(result_name, url, result, metadata)
            if table_name is not None:
                self.results.set_results(
                    table_name=table_name,
                    result=result,
                    result_name=result_name,
                    metadata=metadata,
                )

    def _get_metadata(
        self, url: str, result_name: Results, job_name: str
    ) -> dict[str, Any] | None:
        match result_name:
            case Results.FIGURES:
                return self._get_figure_metadata(url, job_name)
            case Results.METADATA:
                return {"kind": job_name}
            case _:
                return None

    def _get_figure_metadata(self, url: str, job_name: str) -> dict[str, Any] | None:
        figures_metadatas = self.file_downloader.download_file(
            self.results_urls[job_name][Results.FIGURES_METADATA][0]
        )
        if isinstance(figures_metadatas, list):
            return self.results.find_figure_metadata(figures_metadatas, url)
        else:
            raise TypeError(
                f"Expected a list, got {type(figures_metadatas)} instead for {url} metadata."
            )

    def get_specific_result(
        self,
        table_name: str,
        job_name: JobKind,
        result: Results = Results.SHUFFLED,
    ) -> TypeResults:
        job_name_str = self.jobs.get_parameters_name(job_name)
        if table_name not in self.config.tables.keys():
            raise ValueError(f"Expected table '{table_name}' to be created.")
        if not self.jobs.has_job(job_name_str):
            raise ValueError(f"Expected job '{job_name}' to be created. Try running it first.")
        if self.results.get_results(table_name, result, job_name_str) is None:
            self._download_specific_result(job_name_str, result)
        return self.results.get_results(table_name, result, job_name_str)

    def get_all_results(self):
        """
        Get all results.

        Returns
        -------
        dict
            A dictionary with the results of each job on every table.
        Each job is a dictionary with the table name as key and the results as value.
        The results are a dictionary with the result name as key and the data as value.
        The data can be a pandas DataFrame or a dictionary depending on the result type.
        """
        self._download_all_files()
        return self.results

    def download_report(self, path: str | None = None, report_type: ReportType = ReportType.BASIC):
        """
        Download the report.

        Parameters
        ----------
        path
            The path to save the report.
        """
        is_pia = report_type == ReportType.PIA
        job_name = self.jobs.get_parameters_name(JobKind.report, pia_report=is_pia)
        if self.results_urls.get(job_name) is None:
            self._retrieve_job_result_urls(job_name)
        report = self.results_urls[job_name][Results.REPORT][0]
        self.file_downloader.download_file(report, path=path)

    def print_parameters(self, table_name: str | None = None) -> None:
        """Print the parameters for a table.

        Parameters
        ----------
        table_name
            The name of the table.
            If None, all parameters will be printed.
        """
        if table_name is None:
            for table_name in self.config.tables.keys():
                self.print_parameters(table_name)
            return
        if table_name not in self.config.tables.keys():
            raise ValueError(f"Expected table '{table_name}' to be created. Try running it first.")

        # Print avatarization parameters
        avatarization_open_dp = getattr(self.config, "avatarization_open_dp", None)
        avatarization_fast_dp = getattr(self.config, "avatarization_fast_dp", None)
        if self.config.avatarization and table_name in self.config.avatarization:
            print(f"--- Avatarization parameters for {table_name}: ---")  # noqa: T201
            print(asdict(self.config.avatarization[table_name]))  # noqa: T201
            print("\n")  # noqa: T201
        elif avatarization_open_dp and table_name in avatarization_open_dp:
            print(f"--- Avatarization OpenDP parameters for {table_name}: ---")  # noqa: T201
            print(asdict(avatarization_open_dp[table_name]))  # noqa: T201
            print("\n")  # noqa: T201
        elif avatarization_fast_dp and table_name in avatarization_fast_dp:
            print(f"--- Avatarization FastDP parameters for {table_name}: ---")  # noqa: T201
            print(asdict(avatarization_fast_dp[table_name]))  # noqa: T201
            print("\n")  # noqa: T201
        else:
            print(f"--- No avatarization parameters set for {table_name} ---")  # noqa: T201
            print("\n")  # noqa: T201

        # Print privacy metrics parameters
        if self.config.privacy_metrics and table_name in self.config.privacy_metrics:
            print(f"--- Privacy metrics for {table_name}: ---")  # noqa: T201
            print(asdict(self.config.privacy_metrics[table_name]))  # noqa: T201
            print("\n")  # noqa: T201
        else:
            print(f"--- No privacy metrics parameters set for {table_name} ---")  # noqa: T201
            print("\n")  # noqa: T201

        # Print signal metrics parameters
        if self.config.signal_metrics and table_name in self.config.signal_metrics:
            print(f"--- Signal metrics for {table_name}: ---")  # noqa: T201
            print(asdict(self.config.signal_metrics[table_name]))  # noqa: T201
        else:
            print(f"--- No signal metrics parameters set for {table_name} ---")  # noqa: T201

    def kill(self):
        """Method not implemented yet."""
        pass

    def shuffled(self, table_name: str) -> pd.DataFrame:
        """
        Get the shuffled data.

        Parameters
        ----------
        table_name
            The name of the table to get the shuffled data from.

        Returns
        -------
        pd.DataFrame
            The shuffled data as a pandas DataFrame.
        """
        shuffled = self.get_specific_result(table_name, JobKind.standard, Results.SHUFFLED)
        if not isinstance(shuffled, pd.DataFrame):
            raise TypeError(f"Expected a pd.DataFrame, got {type(shuffled)} instead.")
        return shuffled

    def sensitive_unshuffled(self, table_name: str) -> pd.DataFrame:
        """
        Get the unshuffled data.
        This is sensitive data and should be used with caution.

        Parameters
        ----------
        table_name
            The name of the table to get the unshuffled data from.

        Returns
        -------
        pd.DataFrame
            The unshuffled data as a pandas DataFrame.
        """
        unshuffled = self.get_specific_result(table_name, JobKind.standard, Results.UNSHUFFLED)
        if not isinstance(unshuffled, pd.DataFrame):
            raise TypeError(f"Expected a pd.DataFrame, got {type(unshuffled)} instead.")
        return unshuffled

    def privacy_metrics(self, table_name: str) -> list[dict]:
        """
        Get the privacy metrics.

        Parameters
        ----------
        table_name
            The name of the table to get the privacy metrics from.

        Returns
        -------
        dict
            The privacy metrics as a list of dictionary.
        """
        results = self.get_specific_result(
            table_name, JobKind.privacy_metrics, Results.PRIVACY_METRICS
        )
        if not isinstance(results, list):
            raise TypeError(f"Expected a list, got {type(results)} instead.")
        return results

    def signal_metrics(self, table_name: str) -> list[dict]:
        """
        Get the signal metrics.

        Parameters
        ----------
        table_name
            The name of the table to get the signal metrics from.

        Returns
        -------
        dict
            The signal metrics as a list of dictionary.
        """
        results = self.get_specific_result(
            table_name, JobKind.signal_metrics, Results.SIGNAL_METRICS
        )
        if not isinstance(results, list):
            raise TypeError(f"Expected a list, got {type(results)} instead.")
        return results

    def render_plot(self, table_name: str, plot_kind: PlotKind, open_in_browser: bool = False):
        """
        Render a plot for a given table.
        The different plot kinds are defined in the PlotKind enum.

        Parameters
        ----------
        table_name
            The name of the table to get the plot from.
        plot_kind
            The kind of plot to render.
        open_in_browser
            Whether to save the plot to a file and open it in a browser.
        """
        results = self.get_specific_result(table_name, JobKind.standard, Results.FIGURES)
        if not isinstance(results, dict):
            raise TypeError(f"Expected a dict, got {type(results)} instead.")
        if plot_kind not in results:
            raise ValueError(f"No {plot_kind} found for table '{table_name}'.")
        plots = results[plot_kind]
        for idx, plot in enumerate(plots):
            filename = None
            if open_in_browser:
                filename = f"{table_name}_{plot_kind.value}_{idx}.html"
                self._save_file(plot, filename=filename)
            self._open_plot(plot, filename=filename)

    def render_privacy_metrics_summary(self, open_in_browser: bool = False):
        """Render the privacy metrics summary table, this table is only generated
        when there is multiple tables."""
        if self.results.get_results("summary", Results.FIGURES, "privacy_metrics") is None:
            self._download_specific_result("privacy_metrics", Results.FIGURES)
        plots = self.results.get_results("summary", Results.FIGURES, "privacy_metrics")
        if not isinstance(plots, dict):
            raise TypeError(f"Expected a dict, got {type(plots)} instead.")
        if PlotKind.METRICS_SUMMARY not in plots:
            raise ValueError(f"No {PlotKind.METRICS_SUMMARY} found.")
        plot = plots[PlotKind.METRICS_SUMMARY][0]
        filename = None
        if open_in_browser:
            filename = f"{PlotKind.METRICS_SUMMARY.value}.html"
            self._save_file(plot, filename=filename)
        self._open_plot(plot, filename=filename)

    def projections(self, table_name: str) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Get the projections.

        Parameters
        ----------
        table_name
            The name of the table to get the projections from.

        Returns
        -------
        pd.DataFrame
            The projections as a pandas DataFrame.
        """
        original_coordinates = self.get_specific_result(
            table_name, JobKind.standard, Results.PROJECTIONS_ORIGINAL
        )
        avatars_coordinates = self.get_specific_result(
            table_name, JobKind.standard, Results.PROJECTIONS_AVATARS
        )
        if not (
            isinstance(avatars_coordinates, pd.DataFrame)
            and isinstance(original_coordinates, pd.DataFrame)
        ):
            raise TypeError(
                "Expected a pd.DataFrame, "
                f"got {type(original_coordinates)} and {type(avatars_coordinates)} instead."
            )
        return original_coordinates, avatars_coordinates

    def table_summary(self, table_name: str) -> pd.DataFrame:
        """
        Get the table summary.

        Parameters
        ----------
        table_name
            The name of the table to get the summary from.

        Returns
        -------
        pd.DataFrame
            The table summary as a dataframe.
        """
        results = self.get_specific_result(table_name, JobKind.advice, Results.ADVICE)
        if not isinstance(results, dict) or "summary" not in results.keys():
            raise ValueError(f"Summary not found in results for table '{table_name}'")

        summary_data = results["summary"]
        if "stats" not in summary_data.keys():
            raise ValueError(f"Summary not found in results for table '{table_name}'")
        return pd.DataFrame(summary_data["stats"])

    def from_yaml(self, yaml_path: str) -> None:
        """Load configuration from a YAML file.

        This replaces the current runner's configuration with the one from the file.

        **Note**: Table data files are not automatically loaded. Use ``upload_file()``
        to provide data before running jobs.

        Parameters
        ----------
        yaml_path : str
            The path to the YAML configuration file.
        """
        with open(yaml_path, "r") as file:
            yaml_string = file.read()

        # Parse YAML using ConfigParser
        self.config = Config.from_yaml(yaml_string)

    def _open_plot(self, plot_html: HTML, filename: str | None = None):
        """Render a plot, optionally saving it and opening it in a browser."""
        if filename:
            file_path = os.path.abspath(filename)
            webbrowser.open(f"file://{file_path}")
        else:
            display(plot_html)

    def _save_file(self, file_content: HTML, filename: str | None = None):
        """Save the HTML file content to a specified path."""
        if filename is None:
            return None
        with open(filename, "w", encoding="utf-8") as file:
            file.write(file_content.data)
